//	BlobGlobalData.h - BlobVaders global data.
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2012  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef BLOBGLOBALDATA_H_
#define BLOBGLOBALDATA_H_

#include "VaderFactory.h"
#include "GameCounter.h"
#include "GlobalData.h"
#include "TwoFloats.h"

class Planet;
class Vader;
class Fleet;

#include <vector>

///	BlobVaders global data.
class BlobGlobalData : public GlobalData
{
  public:
	///	Constructor.
	BlobGlobalData();
	///	Destructor.
	~BlobGlobalData();

	///	Returns our GameCounter.
	GameCounter& getGameCounter() {return gameCounter;};

	///	Sets the flashing background.
	void setFlashingBackground(bool val);
	///	Sets the screen shake.
	void setScreenShake(bool val);

	///	Returns the flashing background.
	bool getFlashingBackground() const {return flash;};
	///	Returns the screen shake.
	bool getScreenShake() const {return shake;};

	///	Resets the data specific to a single playthrough.
	void resetGame();
	///	Increments the number of fleets the player has defeated.
	/*!
		Will also remove fleets[currentFleetIndex] from the vector.
	 */
	void incNumFleetsDefeated();
	///	Sets our copy of the planets in the solar system.
	void storePlanets(const std::vector<shared_ptr<Planet> >& planets);
	///	Sets our copy of the fleets in the solar system.
	void storeFleets(const std::vector<shared_ptr<Fleet> >& fleets);
	///	Sets the index of the fleet the player is currently fighting.
	void setCurrentFleetIndex(int index);
	///	Sets the player's position within the solar system.
	void setPlayerPosition(const TwoFloats& pos);

	///	Increments the game's difficulty.
	void incrementDifficulty();

	///	Returns the number of fleets the player has defeated.
	int getNumFleetsDefeated() const {return numFleetsDefeated;};
	///	Returns our copy of the planets in the solar system.
	const std::vector<shared_ptr<Planet> >& getPlanets() const {return storedPlanets;};
	///	Returns our copy of the fleets in the solar system.
	const std::vector<shared_ptr<Fleet> >& getFleets() const {return storedFleets;};
	///	Returns the player's position within the solar system.
	const TwoFloats& getPlayerPosition() const {return playerPosition;};

	///	Returns the game's VaderFactory.
	const VaderFactory& getVaderFactory() const {return vaderFactory;};
	///	Returns the game's current difficulty level.
	int getDifficulty() const {return difficulty;};

	///	Adds a vader to the encyclopedia.
	void addToEncyclopedia(const shared_ptr<Vader>& vader);
	///	Returns the number of vaders in the encyclopedia.
	int getEncyclopediaSize() const {return vaderEncyclopedia.size();};
	///	Returns the indexed vader in the encyclopedia.
	const shared_ptr<Vader>& getEncyclopediaVader(int index) const {return vaderEncyclopedia[index];};
  private:
	///	Our GameCounter.
	GameCounter gameCounter;

	///	Whether or not the background flashes.
	bool flash;
	///	Whether or not the screen shakes.
	bool shake;

	///	The number of fleets the player has defeated.
	int numFleetsDefeated;
	///	The planets in the solar system.
	std::vector<shared_ptr<Planet> > storedPlanets;
	///	The fleets in the solar system.
	std::vector<shared_ptr<Fleet> > storedFleets;
	///	The index of the fleet the player is currently fighting against.
	int currentFleetIndex;
	///	The player's position within the solar system.
	TwoFloats playerPosition;

	///	The game's VaderFactory.
	VaderFactory vaderFactory;
	///	The game's current difficulty level.
	int difficulty;

	///	The vaders the player has encountered so far.
	std::vector<shared_ptr<Vader> > vaderEncyclopedia;
};

#endif
