//	BlobVaderSounder.h - The game's sounder.
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2012  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef BLOBVADERSOUNDER_H_
#define BLOBVADERSOUNDER_H_

#include "synth/BrushStrokesSynth.h"
#include "synth/PolyphonicSynth.h"
#include "synth/Synthesizer.h"
#include "synth/SimpleSynth.h"
#include "SequencerCursor.h"
#include "sound/Sounder.h"
#include "sound/Delay.h"
#include "BeatingSine.h"
#include "c0xHeaders.h"

///	The game's sounder.
class BlobVaderSounder : public Sounder,
						 public SequencerCursor::Callback
{
  public:
	///	Constructor.
	BlobVaderSounder();
	///	Destructor.
	~BlobVaderSounder();

	///	Called for every audio block.
	/*!
		This is where you do any audio generation/playback/calculations.
	 */
	void getAudio(float *input, float *output, int numSamples);

	///	Called for each sound file resource by LoadingScreen.
	void loadSoundFileResource(SoundFileLoaderPtr soundFileLoader) {};

	///	Called to start the sequence going.
	void startSequence();
	///	Called to stop the sequence.
	void stopSequence();
	///	Returns whether or not the sequence is currently playing.
	bool sequenceIsPlaying() const {return playing;};

	///	Sets the amount of action on screen.
	void setAmountAction(float val);
	///	Adds a note to the sequencer.
	void addNote(int synth, int cell, int freq);
	///	Sets the indexed pad step.
	void setPadStep(int cell, int freq);

	///	Triggers an SFX synth.
	void triggerSFX(int synth);

	///	Sets the amplitude of the indexed BeatingSine.
	/*!
		Note: also sets the frequency of the brush strokes synth.
	 */
	void setSineAmplitude(int index, float val);
	///	Activates/deactivates the brush strokes synth.
	void activateBrushStrokes(bool val);
	///	Updates the brush strokes synth's trail.
	void setBrushStrokesTrail(TwoFloats *trail);
	///	Turns the sun synth on/off.
	void activateSun(bool val);
	///	Activates or deactivates a fleet synth.
	void activateFleet(int index, bool val);

	///	Called whenever the cursor moves to the next step in the sequence.
	void nextStepCallback(int step, int sample);

	///	Helper method. Selects a random pitch set from the pitches dir.
	void loadPitches();
	///	Loads a new set of pitches for the planet sines.
	void loadPlanetPitches();

	///	Callback used to advance the BlobVaders.
	class Callback
	{
	  public:
		///	Destructor.
		virtual ~Callback() {};

		///	Called for every beat.
		virtual void beatCallback() = 0;
	};
	///	Registers a Callback.
	void registerCallback(Callback *c);
  private:
	///	The number of steps in the sequencer, etc.
	enum
	{
		NumSteps = 16,
		Kick = 0,
		Clap,
		Hat,
		Crash,

		NumSines = 6
	};

	///	The sequence's cursor.
	SequencerCursor cursor;
	
	///	The synth sequences.
	float steps[2][NumSteps];
	///	The pad synth sequence.
	float padSteps[4];
	///	The current pad step.
	int currentPadStep;
	///	Used to only add new notes once old ones have decayed.
	float stepDecays[2][NumSteps];
	///	The 3 synths.
	PolyphonicSynth<SimpleSynth> synths[3];
	///	The 4 drum sounds.
	shared_ptr<Synthesizer> drums[4];

	///	The 3 SFX synths.
	PolyphonicSynth<SimpleSynth> sfxSynths[3];

	///	The beating sines used in the solar system.
	BeatingSine planetSines[NumSines];
	///	The brush strokes synth used in the solar system.
	BrushStrokesSynth brushStrokes;
	///	Delay used on the brush strokes synth.
	Delay brushDelay;
	///	The buffer used by the brush strokes synth.
	float brushBuffer[88200];
	///	The sun synth.
	SimpleSynth sunSynth;
	///	The fleet synths.
	SimpleSynth fleetSynths[NumSines*2];

	///	The delay on the high synth.
	Delay hiDelay;
	///	The buffer used by the high synth.
	float delayBuffer[88200];

	///	Whether the sequence is currently playing or not.
	bool playing;

	///	Used to determine when to add hats and crashes.
	float action;

	///	The current pitch set.
	float pitches[8];
	///	The pitches used by the planets.
	float planetPitches[8];

	///	Our callback.
	Callback *callback;
};

#endif
