//	GamePlayState.cpp - The main gameplay state.
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2012  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "Random.h"
#include "BlobVader.h"
#include "VaderBullet.h"
#include "Application.h"
#include "PlayerBullet.h"
#include "GamePlayState.h"
#include "drawing/Drawer.h"
#include "BlobGlobalData.h"
#include "BlobVaderSounder.h"

#include <sstream>

using namespace std;

//------------------------------------------------------------------------------
GameStateCreator<GamePlayState> GamePlayState::creator;

//------------------------------------------------------------------------------
GamePlayState::GamePlayState(Application& app, SounderPtr& sound):
GameState(app, sound),
backgroundColour(0.1f),
backgroundTweener(backgroundColour, ThreeFloats(0.1f), 0.1f),
blobColourLeft(1.0f, 0.0f, 0.0f),
blobLeftTweener(blobColourLeft, ThreeFloats(0.0f, 0.0f, 1.0f), 0.01f),
blobColourRight(0.0f, 0.0f, 1.0f),
blobRightTweener(blobColourRight, ThreeFloats(0.0f, 1.0f, 0.0f), 0.01f),
blobMoveIndex(0),
blobMoveCount(120),
blobMove(false),
vaderBulletCount(120),
vaderBulletInc(1),
shake(0.0f),
shakeIndex(0.0f),
gameOver(false),
gameWin(false),
gameWinText(L"Fleet Defeated!\n\n\n\n\n<click to return>"),
musicAction(0.0f),
notePos(0),
killDistance(0.0f),
timeElapsed(0.0f),
sounder(dynamic_pointer_cast<BlobVaderSounder>(sound)),
endPause(0)
{
	int x, y;

	lastTicks = getApplication().getTimeSinceStart();

	globalData = dynamic_pointer_cast<BlobGlobalData>(getGlobalData());
	globalData->getGameCounter().registerNotification(L"Dead Vader", 60, this);
	backgroundFlash = globalData->getFlashingBackground();
	screenShake = globalData->getScreenShake();

	remainingWaves = 10 + (globalData->getNumFleetsDefeated() * 4);

	for(y=0;y<4;++y)
	{
		for(x=0;x<7;++x)
		{
			blobVaders.push_back(BlobVaderPtr(new BlobVader(TwoFloats(128.0f+((1280.0f-256.0f) * ((float)x/8.0f)),
															(72.0f+((640-256.0f) * ((float)y/4.0f)))-448.0f))));
		}
	}

	sounder->registerCallback(this);
	sounder->loadPitches();
}

//------------------------------------------------------------------------------
GamePlayState::~GamePlayState()
{
	sounder->setAmountAction(0.0f);
	sounder->stopSequence();
	globalData->getGameCounter().unregisterNotification(L"Dead Vader", 60, this);
}

//------------------------------------------------------------------------------
void GamePlayState::update()
{
	int i;
	bool onBottomRow = false;

	if(!gameOver && !gameWin)
	{
		const uint32_t ticks = getApplication().getTimeSinceStart();

		timeElapsed += (float)(ticks-lastTicks) * 0.001f;
		lastTicks = ticks;

		//Step the BlobVaders across the screen.
		--blobMoveCount;
		if(!blobMoveCount)
		{
			if(!sounder->sequenceIsPlaying())
				sounder->startSequence();

			blobMoveCount = 60;
		}

		if(blobMove)
		{
			blobMove = false;

			for(i=0;i<(int)blobVaders.size();++i)
			{
				if(blobVaders[i]->getPosition().y >= 600.0f)
				{
					onBottomRow = true;
					break;
				}
			}

			++blobMoveIndex;
			if(blobMoveIndex == 3)
			{
				if(onBottomRow)
				{
					++blobMoveIndex;
					for(i=0;i<(int)blobVaders.size();++i)
					{
						if(!blobVaders[i]->getDead() && !blobVaders[i]->getInvulnerable())
							blobVaders[i]->stepLeft();
					}
				}
				else
				{
					for(i=0;i<(int)blobVaders.size();++i)
					{
						if(!blobVaders[i]->getDead() && !blobVaders[i]->getInvulnerable())
							blobVaders[i]->stepDown();
					}
					addNewVaders();
				}
			}
			else if(blobMoveIndex == 6)
			{
				if(onBottomRow)
				{
					blobMoveIndex = 1;
					for(i=0;i<(int)blobVaders.size();++i)
					{
						if(!blobVaders[i]->getDead() && !blobVaders[i]->getInvulnerable())
							blobVaders[i]->stepRight();
					}
				}
				else
				{
					blobMoveIndex = 0;
					for(i=0;i<(int)blobVaders.size();++i)
					{
						if(!blobVaders[i]->getDead() && !blobVaders[i]->getInvulnerable())
							blobVaders[i]->stepDown();
					}
					addNewVaders();
				}
			}
			else if(blobMoveIndex > 3)
			{
				for(i=0;i<(int)blobVaders.size();++i)
				{
					if(!blobVaders[i]->getDead() && !blobVaders[i]->getInvulnerable())
						blobVaders[i]->stepLeft();
				}
			}
			else
			{
				for(i=0;i<(int)blobVaders.size();++i)
				{
					if(!blobVaders[i]->getDead() && !blobVaders[i]->getInvulnerable())
						blobVaders[i]->stepRight();
				}
			}
		}

		//Update the BlobVaders.
		for(i=0;i<(int)blobVaders.size();++i)
			blobVaders[i]->update();

		//Add any new VaderBullets.
		vaderBulletCount -= vaderBulletInc;
		if(vaderBulletCount < 0)
		{
			while(vaderBulletCount < 0)
			{
				int index = Random::getGlobal().getRandomInt(blobVaders.size()-1);
				vaderBulletCount += 120 + ((49-blobVaders.size())*16);

				if(!blobVaders[index]->getDead())
				{
					vaderBullets.push_back(VaderBulletPtr(new VaderBullet(blobVaders[index]->getPosition())));
					blobVaders[index]->scaleUp();

					sounder->triggerSFX(1);
				}
			}
			if(vaderBulletInc < 30)
				++vaderBulletInc;
		}

		//Delete any finished BlobVaders.
		for(i=(blobVaders.size()-1);i>=0;--i)
		{
			if(blobVaders[i]->getFinished())
				blobVaders.erase(blobVaders.begin()+i);
		}

		//Update the PlayerBullets.
		for(i=(playerBullets.size()-1);i>=0;--i)
		{
			if(playerBullets[i]->update())
			{
				shake += 1.0f;
				if(shake > 2.0f)
					shake = 2.0f;
				particles.trigger(playerBullets[i]->getPosition());

				sounder->triggerSFX(2);
			}

			if(playerBullets[i]->getPosition().y < -360.0f)
				playerBullets.erase(playerBullets.begin()+i);
			else if(playerBullets[i]->getPosition().y > 800.0f)
				playerBullets.erase(playerBullets.begin()+i);
		}

		//Update the VaderBullets.
		for(i=(vaderBullets.size()-1);i>=0;--i)
		{
			vaderBullets[i]->update();

			if(vaderBullets[i]->getPosition().y > 900.0f)
				vaderBullets.erase(vaderBullets.begin()+i);
		}

		//Update the player ship.
		player.update();

		//Handle PlayerBullet->BlobVader collisions.
		collideBulletsVaders();
		//Handle VaderBullet->Player collisions.
		collideBulletsPlayer();
		//Handle BlobVader->Player collisions.
		collideVadersPlayer();

		if(!remainingWaves && !blobVaders.size())
		{
			gameWin = true;
			endPause = 60;
		}
	}
	if(endPause)
		--endPause;

	//Update any particles.
	particles.update();

	//Calculate the amount of action for the sounder.
	if(musicAction > 0.0f)
	{
		musicAction *= 0.5f;
		if(musicAction < 0.1f)
			musicAction = 0.0f;
	}
	musicAction += ((float)blobVaders.size()/192.0f);
	musicAction += ((float)playerBullets.size()/72.0f);
	musicAction += ((float)vaderBullets.size()/72.0f);
	musicAction += shake * 0.25f;
	if(musicAction > 1.0f)
		musicAction = 1.0f;
	sounder->setAmountAction(musicAction);

	//Update killDistance.
	if(notePos > 0)
		killDistance += 0.2f;
}

//------------------------------------------------------------------------------
void GamePlayState::draw(DrawerPtr& d)
{
	unsigned int i;
	TwoFloats shakeVal;
	ThreeFloats tempCol;
	const TwoFloats origin;

	if(!backgroundTweener.getFinished() && backgroundFlash)
		backgroundTweener.tick();
	/*else
	{
		ThreeFloats newCol;

		newCol.x = Random::getGlobal().getRandomFloat() * 0.33f;
		newCol.y = Random::getGlobal().getRandomFloat() * 0.33f;
		newCol.z = Random::getGlobal().getRandomFloat() * 0.33f;
		backgroundTweener.setDelta(0.01f);
		backgroundTweener.reset(newCol);
	}*/
	d->fillRect(origin, d->getDrawingAreaSize(), backgroundColour);
	//d->fillRect(origin, d->getDrawingAreaSize(), ThreeFloats(0.1f));

	if(!blobRightTweener.getFinished())
		blobRightTweener.tick();
	else
	{
		ThreeFloats tempCol;

		while((tempCol == 0.0f) || (tempCol == 1.0f) || (tempCol == blobColourRight))
		{
			tempCol.x = (Random::getGlobal().getRandomFloat() > 0.5f) ? 1.0f : 0.0f;
			tempCol.y = (Random::getGlobal().getRandomFloat() > 0.5f) ? 1.0f : 0.0f;
			tempCol.z = (Random::getGlobal().getRandomFloat() > 0.5f) ? 1.0f : 0.0f;
		}
		blobRightTweener.reset(tempCol);
	}

	if(!blobLeftTweener.getFinished())
		blobLeftTweener.tick();
	else
	{
		blobLeftTweener.reset(blobColourRight);
	}

	if((shake > 0.0f) && screenShake)
	{
		//0.75f is the toned-down value.
		shakeVal.x = sinf(shakeIndex*2.0f*NIALL_PI) * 16.0f * shake * 0.75f;
		d->translate(shakeVal);
		d->rotate(shakeVal.x * 0.001f, TwoFloats(640.0f, 360.0f));

		shake -= 0.05f;
		shakeIndex += 0.1f;
		if(shakeIndex > 1.0f)
			shakeIndex -= 1.0f;
	}

	for(i=0;i<vaderBullets.size();++i)
		vaderBullets[i]->draw(d);

	//d->setBlendMode(Drawer::Additive);
	for(i=0;i<blobVaders.size();++i)
		blobVaders[i]->draw(d, blobColourLeft, blobColourRight);
	//d->setBlendMode(Drawer::AlphaBlend);

	for(i=0;i<playerBullets.size();++i)
		playerBullets[i]->draw(d);

	tempCol.x = 1.0f-blobColourLeft.x;
	tempCol.y = 1.0f-blobColourLeft.y;
	tempCol.z = 1.0f-blobColourLeft.z;
	player.draw(d, tempCol);

	particles.draw(d);

	//Draw the score.
	if(0)
	{
		wstringstream tempstr;

		tempstr << L"Score: " << globalData->getGameCounter().getScore();

		d->drawText(tempstr.str(),
					L"Font0",
					TwoFloats(10.0f, 44.0f),
					384.0f,
					TextLayout::Left,
					1.0f,
					ThreeFloats(1.0f));
	}

	if((shake > 0.0f) && screenShake)
	{
		shakeVal.x = -shakeVal.x;
		d->translate(shakeVal);
		d->rotate(-shakeVal.x * 0.001f, TwoFloats(640.0f, 360.0f));
	}

	if(gameOver)
	{
		d->drawText(gameOverText,
					L"Font0",
					gameOverLayout,
					TwoFloats(0.0f, 320.0f),
					d->getDrawingAreaSize().x,
					TextLayout::Centre,
					1.0f,
					ThreeFloats(1.0f));
	}
	else if(gameWin)
	{
		d->drawText(gameWinText,
					L"Font0",
					gameWinLayout,
					TwoFloats(0.0f, 320.0f),
					d->getDrawingAreaSize().x,
					TextLayout::Centre,
					1.0f,
					ThreeFloats(1.0f));
	}
}

//------------------------------------------------------------------------------
void GamePlayState::keyDown(const wstring& action,
							const wstring& keyPressed,
							int keyModifier)
{
	if(action == L"Escape")
		nextState(L"TitleScreen");
	/*else if(action == L"Action")
	{
		if(gameOver)
		{
			int x, y;

			gameOver = false;

			blobMoveIndex = 0;
			blobMoveCount = 120;
			blobMove = false;
			vaderBulletCount = 120;
			vaderBulletInc = 1;
			globalData->getGameCounter().reset();
			gameOverLayout.clear();

			blobVaders.clear();
			vaderBullets.clear();
			playerBullets.clear();

			for(y=0;y<4;++y)
			{
				for(x=0;x<7;++x)
				{
					blobVaders.push_back(BlobVaderPtr(new BlobVader(TwoFloats(128.0f+((1280.0f-256.0f) * ((float)x/8.0f)),
																	(72.0f+((640-256.0f) * ((float)y/4.0f)))-448.0f))));
				}
			}

			musicAction = 0.0f;
			notePos = 0;
			killDistance = 0.0f;
			sounder->setAmountAction(0.0f);
			sounder->stopSequence();
			sounder->loadPitches();
		}
	}*/
}

//------------------------------------------------------------------------------
void GamePlayState::keyUp(const wstring& action,
							const wstring& keyPressed,
							int keyModifier)
{
	
}

//------------------------------------------------------------------------------
void GamePlayState::mouseDown(int button, const TwoFloats& pos)
{
	if(gameWin && !endPause)
	{
		globalData->incNumFleetsDefeated();
		nextState(L"SolarSystemState");
		return;
	}
	else if(!gameOver)
	{
		TwoFloats dir;
		ThreeFloats tempCol;
		float speed = player.getSpeed();

		dir.x = sinf(speed * 0.2f);
		dir.y = -cosf(speed * 0.2f);

		playerBullets.push_back(PlayerBulletPtr(new PlayerBullet(player.getPosition(), dir)));
		player.fired();

		sounder->triggerSFX(0);
	}
	else if(!endPause)
	{
		/*int x, y;

		gameOver = false;

		blobMoveIndex = 0;
		blobMoveCount = 120;
		blobMove = false;
		vaderBulletCount = 120;
		vaderBulletInc = 1;
		globalData->getGameCounter().reset();
		gameOverLayout.clear();

		blobVaders.clear();
		vaderBullets.clear();
		playerBullets.clear();

		for(y=0;y<4;++y)
		{
			for(x=0;x<7;++x)
			{
				blobVaders.push_back(BlobVaderPtr(new BlobVader(TwoFloats(128.0f+((1280.0f-256.0f) * ((float)x/8.0f)),
																(72.0f+((640-256.0f) * ((float)y/4.0f)))-448.0f))));
			}
		}

		musicAction = 0.0f;
		notePos = 0;
		killDistance = 0.0f;
		sounder->setAmountAction(0.0f);
		sounder->stopSequence();
		sounder->loadPitches();*/
		nextState(L"TitleScreen");
		return;
	}
}

//------------------------------------------------------------------------------
void GamePlayState::mouseMove(const TwoFloats& pos)
{
	player.mouseMove(pos);
}

//------------------------------------------------------------------------------
void GamePlayState::beatCallback()
{
	blobMove = true;
}

//------------------------------------------------------------------------------
void GamePlayState::countReached(const wstring& counter, int value)
{
	int i, j;
	int tempint;
	float tempf;
	int startVaders = 0;
	const TwoFloats origin;
	vector<int> availableFreqs;

	if(counter == L"Dead Vader")
	{
		if(musicAction > 0.33f)
		{
			for(i=0;i<8;++i)
				availableFreqs.push_back(i);

			for(i=0;i<4;++i)
			{
				tempf = 0.0f;
				for(j=startVaders;j<(startVaders+((int)blobVaders.size()/4));++j)
				{
					tempf += origin.getDistance(blobVaders[j]->getPosition());
				}
				startVaders += (blobVaders.size()/4);

				tempf *= 0.25f;
				tempf /= 1468.60478f;
				tempf *= ((float)availableFreqs.size())-0.01f;
				tempint = (int)tempf;

				sounder->setPadStep(i, availableFreqs[tempint]);

				availableFreqs.erase(availableFreqs.begin()+tempint);
			}
		}
		else
		{
			for(i=0;i<4;++i)
				sounder->setPadStep(i, -1);
		}
	}
}

//------------------------------------------------------------------------------
void GamePlayState::collideBulletsVaders()
{
	int i, j;
	TwoFloats bulletPos;
	TwoFloats vaderPos;
	bool killBullet;

	for(i=(playerBullets.size()-1);i>=0;--i)
	{
		killBullet = false;
		bulletPos = playerBullets[i]->getPosition();

		if(bulletPos.y < 0.0f)
			continue;

		for(j=0;j<(int)blobVaders.size();++j)
		{
			if(!blobVaders[j]->getDead() && !blobVaders[j]->getInvulnerable())
			{
				vaderPos = blobVaders[j]->getPosition();

				if(bulletPos.getDistance(vaderPos) < 40.0f)
				{
					int colourIndex;

					killBullet = true;

					blobVaders[j]->setDead();

					shake += 1.0f;
					if(shake > 2.0f)
						shake = 2.0f;
					particles.trigger(blobVaders[j]->getPosition());
					globalData->getGameCounter().incrementCount(L"Dead Vader");
					
					if(backgroundFlash)
					{
						colourIndex = Random::getGlobal().getRandomInt(3);
						switch(colourIndex)
						{
							case 0:
								backgroundColour.x += 0.25f;
								break;
							case 1:
								backgroundColour.y += 0.25f;
								break;
							case 2:
								backgroundColour.z += 0.25f;
								break;
							default:
								backgroundColour.x += 0.25f;
						}
						backgroundTweener.reset(ThreeFloats(0.1f));
					}

					addNote(blobVaders[j]->getPosition());

					break;
				}
			}
		}
		if(killBullet)
			playerBullets.erase(playerBullets.begin()+i);
	}
}

//------------------------------------------------------------------------------
void GamePlayState::collideBulletsPlayer()
{
	int i;
	TwoFloats bulletPos;
	const TwoFloats& playerPos = player.getPosition();

	for(i=(vaderBullets.size()-1);i>=0;--i)
	{
		bulletPos = vaderBullets[i]->getPosition();
		if(bulletPos.getDistance(playerPos) < 32.0f)
		{
			wstringstream tempstr;

			particles.trigger(vaderBullets[i]->getPosition());

			vaderBullets.erase(vaderBullets.begin()+i);

			endPause = 60;
			gameOver = true;
			tempstr << L"GAME OVER\n\nVaders Killed: " << globalData->getGameCounter().getScore();
			tempstr << L"\n\n\n\n\n<click to title>";
			gameOverText = tempstr.str();

			break;
		}
	}
}

//------------------------------------------------------------------------------
void GamePlayState::collideVadersPlayer()
{
	int i;
	int colourIndex;
	TwoFloats bulletPos;
	const TwoFloats& playerPos = player.getPosition();

	for(i=(blobVaders.size()-1);i>=0;--i)
	{
		bulletPos = blobVaders[i]->getPosition();
		if(bulletPos.getDistance(playerPos) < 42.0f)
		{
			wstringstream tempstr;

			particles.trigger(blobVaders[i]->getPosition());

			blobVaders[i]->setDead();

			shake += 1.0f;
			if(shake > 2.0f)
				shake = 2.0f;
			particles.trigger(blobVaders[i]->getPosition());
			globalData->getGameCounter().incrementCount(L"Dead Vader");

			colourIndex = Random::getGlobal().getRandomInt(3);
			switch(colourIndex)
			{
				case 0:
					backgroundColour.x += 0.25f;
					break;
				case 1:
					backgroundColour.y += 0.25f;
					break;
				case 2:
					backgroundColour.z += 0.25f;
					break;
				default:
					backgroundColour.x += 0.25f;
			}
			backgroundTweener.reset(ThreeFloats(0.1f));

			endPause = 60;
			gameOver = true;
			tempstr << L"GAME OVER\n\nVaders Killed: " << globalData->getGameCounter().getScore();
			tempstr << L"\n\n\n\n\n<click to title>";
			gameOverText = tempstr.str();

			break;
		}
	}
}

//------------------------------------------------------------------------------
void GamePlayState::addNewVaders()
{
	int x;

	if(remainingWaves > 0)
	{
		for(x=1;x<8;++x)
		{
			blobVaders.push_back(BlobVaderPtr(new BlobVader(TwoFloats(128.0f+((1280.0f-256.0f) * ((float)x/8.0f)),
															(72.0f+((640-256.0f) * ((float)0/4.0f)))-448.0f))));
		}
		--remainingWaves;
	}
}

//------------------------------------------------------------------------------
void GamePlayState::addNote(const TwoFloats& pos)
{
	int note;
	int cell;
	int synth = 0;
	const TwoFloats origin;
	const float maxDist = 1468.60478f;

	cell = (notePos + (int)killDistance + (int)((1.0f-musicAction) * 4.0f)) % 16;
	note = (int)((origin.getDistance(pos)/maxDist) * 15.99f);
	if(note > 7)
	{
		synth = 1;
		note -= 8;
	}

	sounder->addNote(synth, cell, note);
	if(synth && (Random::getGlobal().getRandomFloat() > 0.5f))
	{
		cell += 8;
		cell %= 16;
		sounder->addNote(synth, cell, note);
	}

	notePos = (cell+1)%16;
	killDistance = 0.0f;
}
