//	NewGamePlayState.cpp - The main gameplay state.
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2012  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "BlobVaderSounder.h"
#include "NewGamePlayState.h"
#include "VaderProjectile.h"
#include "VaderExplosion.h"
#include "drawing/Drawer.h"
#include "BlobGlobalData.h"
#include "PlayerBullet.h"
#include "Application.h"
#include "Random.h"
#include "Vader.h"

#include <sstream>

using namespace std;

//------------------------------------------------------------------------------
GameStateCreator<NewGamePlayState> NewGamePlayState::creator;

//------------------------------------------------------------------------------
NewGamePlayState::NewGamePlayState(Application& app, SounderPtr& sound):
GameState(app, sound),
backgroundColour(0.1f),
backgroundTweener(backgroundColour, ThreeFloats(0.1f), 0.1f),
blobColourLeft(1.0f, 0.0f, 0.0f),
blobLeftTweener(blobColourLeft, ThreeFloats(0.0f, 0.0f, 1.0f), 0.01f),
blobColourRight(0.0f, 0.0f, 1.0f),
blobRightTweener(blobColourRight, ThreeFloats(0.0f, 1.0f, 0.0f), 0.01f),
blobMoveIndex(0),
blobMoveCount(120),
blobMove(false),
vaderBulletCount(120),
vaderBulletInc(1),
vaderBulletSpeed(1.0f),
player(new Player()),
shake(0.0f),
shakeIndex(0.0f),
gameOver(false),
gameWin(false),
gameWinText(L"Fleet Defeated!\n\n\n\n\n<click to return>"),
musicAction(0.0f),
notePos(0),
killDistance(0.0f),
timeElapsed(0.0f),
autofire(false),
autofireCount(0),
autofireMax(15),
sounder(dynamic_pointer_cast<BlobVaderSounder>(sound)),
globalData(dynamic_pointer_cast<BlobGlobalData>(getGlobalData())),
endPause(0)
{
	int i, j;
	int index;
	bool alreadyUsed;
	TwoFloats tempPos;
	int numVaderTypes = 3;
	std::wstring vaderName;
	const VaderFactory& factory = globalData->getVaderFactory();
	std::vector<std::wstring> typesUsed;

	lastTicks = getApplication().getTimeSinceStart();

	globalData->getGameCounter().registerNotification(L"Dead Vader", 60, this);
	backgroundFlash = globalData->getFlashingBackground();
	screenShake = globalData->getScreenShake();

	collidables.push_back(weak_ptr<Collidable>(player));

	remainingWaves = 10 + (globalData->getNumFleetsDefeated() * 4);
	numVaderTypes += (globalData->getDifficulty()-1);
	if(numVaderTypes > 6)
		numVaderTypes = 6;

	//Decide which vader types we're going to use.
	while((int)typesUsed.size() < numVaderTypes) 
	{
		vaderName = factory.getVaderName(globalData->getDifficulty(),
										 Random::getGlobal().getRandomInt(factory.getNumVaderTypes()));

		alreadyUsed = false;
		for(i=0;i<(int)typesUsed.size();++i)
		{
			if(vaderName == typesUsed[i])
			{
				alreadyUsed = true;
				break;
			}
		}
		if(!alreadyUsed)
			typesUsed.push_back(vaderName);
	}
	//Fill out typesUsed so we get a roughly even spread of the different types.
	while((int)typesUsed.size() < remainingWaves)
		typesUsed.push_back(typesUsed[Random::getGlobal().getRandomInt(numVaderTypes)]);

	//Setup the vader waves.
	waveTypes.push_back(factory.getVaderName(0, 0));
	globalData->addToEncyclopedia(factory.createVader(waveTypes.back(),
													  TwoFloats(0.0f),
													  *this));
	while((int)waveTypes.size() < remainingWaves)
	{
		index = Random::getGlobal().getRandomInt(typesUsed.size());
		waveTypes.push_back(typesUsed[index]);
		typesUsed.erase(typesUsed.begin()+index);

		globalData->addToEncyclopedia(factory.createVader(waveTypes.back(),
														  TwoFloats(0.0f),
														  *this));
	}

	/*for(y=0;y<4;++y)
	{
		for(x=0;x<7;++x)
		{
			blobVaders.push_back(BlobVaderPtr(new BlobVader(TwoFloats(128.0f+((1280.0f-256.0f) * ((float)x/8.0f)),
															(72.0f+((640-256.0f) * ((float)y/4.0f)))-448.0f))));
		}
	}*/
	//Setup the first four vader waves.
	for(i=0;i<4;++i)
	{
		for(j=0;j<7;++j)
		{
			tempPos.x = 128.0f+((getDrawingAreaSize().x-256.0f) * ((float)j/8.0f));
			tempPos.y = (72.0f+((640-256.0f) * ((float)(4-(i+1))/4.0f))) - 448.0f;

			blobVaders.push_back(factory.createVader(waveTypes[0],
													 tempPos,
													 *this));
			collidables.push_back(weak_ptr<Vader>(blobVaders.back()));
		}
		waveTypes.erase(waveTypes.begin());
		--remainingWaves;
	}

	sounder->registerCallback(this);
	sounder->loadPitches();
}

//------------------------------------------------------------------------------
NewGamePlayState::~NewGamePlayState()
{
	sounder->setAmountAction(0.0f);
	sounder->stopSequence();
	globalData->getGameCounter().unregisterNotification(L"Dead Vader", 60, this);
}

//------------------------------------------------------------------------------
void NewGamePlayState::update()
{
	int i;
	bool onBottomRow = false;
	const TwoFloats& playerPos = player->getPosition();

	if(player->getDead() && !gameOver)
	{
		std::wstringstream tempstr;

		endPause = 40;
		gameOver = true;
		tempstr << L"GAME OVER\n\nVaders Killed: " << globalData->getGameCounter().getScore();
		tempstr << L"\n\n\n\n\n<click to title>";
		gameOverText = tempstr.str();
	}

	if(!gameOver && !gameWin)
	{
		const uint32_t ticks = getApplication().getTimeSinceStart();

		timeElapsed += (float)(ticks-lastTicks) * 0.001f;
		lastTicks = ticks;

		vaderBulletSpeed += 0.001f;
		for(i=0;i<(int)blobVaders.size();++i)
			blobVaders[i]->setBulletSpeed((int)vaderBulletSpeed);

		//Add any new PlayerBullets.
		if(autofire)
		{
			--autofireCount;
			if(!autofireCount)
			{
				firePlayerBullet();
				autofireCount = autofireMax;
			}
		}

		//Step the BlobVaders across the screen.
		--blobMoveCount;
		if(!blobMoveCount)
		{
			if(!sounder->sequenceIsPlaying())
				sounder->startSequence();

			blobMoveCount = 60;
		}

		if(blobMove)
		{
			blobMove = false;

			for(i=0;i<(int)blobVaders.size();++i)
			{
				if(blobVaders[i]->getPosition().y >= 600.0f)
				{
					onBottomRow = true;
					break;
				}
			}

			++blobMoveIndex;
			if(blobMoveIndex == 3)
			{
				if(onBottomRow)
				{
					++blobMoveIndex;
					for(i=0;i<(int)blobVaders.size();++i)
					{
						if(!blobVaders[i]->getDead() && !blobVaders[i]->doNotCount())
							blobVaders[i]->stepLeft();
					}
				}
				else
				{
					for(i=0;i<(int)blobVaders.size();++i)
					{
						if(!blobVaders[i]->getDead() && !blobVaders[i]->doNotCount())
							blobVaders[i]->stepDown();
					}
					addNewVaders();
				}
			}
			else if(blobMoveIndex == 6)
			{
				if(onBottomRow)
				{
					blobMoveIndex = 1;
					for(i=0;i<(int)blobVaders.size();++i)
					{
						if(!blobVaders[i]->getDead() && !blobVaders[i]->doNotCount())
							blobVaders[i]->stepRight();
					}
				}
				else
				{
					blobMoveIndex = 0;
					for(i=0;i<(int)blobVaders.size();++i)
					{
						if(!blobVaders[i]->getDead() && !blobVaders[i]->doNotCount())
							blobVaders[i]->stepDown();
					}
					addNewVaders();
				}
			}
			else if(blobMoveIndex > 3)
			{
				for(i=0;i<(int)blobVaders.size();++i)
				{
					if(!blobVaders[i]->getDead() && !blobVaders[i]->doNotCount())
						blobVaders[i]->stepLeft();
				}
			}
			else
			{
				for(i=0;i<(int)blobVaders.size();++i)
				{
					if(!blobVaders[i]->getDead() && !blobVaders[i]->doNotCount())
						blobVaders[i]->stepRight();
				}
			}
		}

		//Update the BlobVaders.
		for(i=0;i<(int)blobVaders.size();++i)
			blobVaders[i]->update();

		//Update the VaderProjectiles.
		for(i=0;i<(int)vaderBullets.size();++i)
			vaderBullets[i]->update(playerPos);

		//Update the PlayerBullets.
		for(i=(playerBullets.size()-1);i>=0;--i)
		{
			if(playerBullets[i]->update())
			{
				shake += 1.0f;
				if(shake > 2.0f)
					shake = 2.0f;
				particles.trigger(playerBullets[i]->getPosition());

				sounder->triggerSFX(2);
			}
		}

		//Update the player ship.
		player->update();

		//Handle collision detection.
		collisionDetection();

		//Delete any finished BlobVaders.
		for(i=(blobVaders.size()-1);i>=0;--i)
		{
			if(blobVaders[i]->getDead())
			{
				shake += 1.0f;
				if(shake > 2.0f)
					shake = 2.0f;

				particles.trigger(blobVaders[i]->getPosition());
				globalData->getGameCounter().incrementCount(L"Dead Vader");

				if(backgroundFlash)
				{
					int colourIndex;

					colourIndex = Random::getGlobal().getRandomInt(3);
					switch(colourIndex)
					{
						case 0:
							backgroundColour.x += 0.25f;
							break;
						case 1:
							backgroundColour.y += 0.25f;
							break;
						case 2:
							backgroundColour.z += 0.25f;
							break;
						default:
							backgroundColour.x += 0.25f;
					}
					backgroundTweener.reset(ThreeFloats(0.1f));
				}

				addNote(blobVaders[i]->getPosition());

				vaderExplosions.push_back(shared_ptr<VaderExplosion>(new VaderExplosion(blobVaders[i]->getImage(),
																						blobVaders[i]->getPosition(),
																						blobVaders[i]->getColour(),
																						blobVaders[i]->getWingMotion())));

				blobVaders.erase(blobVaders.begin()+i);
			}
		}

		//Delete any finished VaderProjectiles.
		for(i=(vaderBullets.size()-1);i>=0;--i)
		{
			if(vaderBullets[i]->getDead())
			{
				vaderBullets[i]->deathAction(*this);

				particles.trigger(vaderBullets[i]->getPosition());
				vaderBullets.erase(vaderBullets.begin()+i);
			}
		}

		//Delete any finished PlayerBullets.
		for(i=(playerBullets.size()-1);i>=0;--i)
		{
			if(playerBullets[i]->getDead())
				playerBullets.erase(playerBullets.begin()+i);
		}

		//Remove any dead entries from collidables.
		for(i=(collidables.size()-1);i>=0;--i)
		{
			if(collidables[i].expired())
				collidables.erase(collidables.begin()+i);
		}

		//Handle win condition.
		if((remainingWaves <= 0) && !blobVaders.size())
		{
			gameWin = true;	
			endPause = 40;
		}
	}
	if(endPause)
		--endPause;

	//Update VaderExplosions.
	for(i=(vaderExplosions.size()-1);i>=0;--i)
	{
		if(vaderExplosions[i]->update())
			vaderExplosions.erase(vaderExplosions.begin()+i);
	}

	//Update any particles.
	particles.update();

	//Calculate the amount of action for the sounder.
	if(musicAction > 0.0f)
	{
		musicAction *= 0.5f;
		if(musicAction < 0.1f)
			musicAction = 0.0f;
	}
	musicAction += ((float)blobVaders.size()/192.0f);
	musicAction += ((float)playerBullets.size()/72.0f);
	musicAction += ((float)vaderBullets.size()/72.0f);
	musicAction += shake * 0.25f;
	if(musicAction > 1.0f)
		musicAction = 1.0f;
	sounder->setAmountAction(musicAction);

	//Update killDistance.
	if(notePos > 0)
		killDistance += 0.2f;
}

//------------------------------------------------------------------------------
void NewGamePlayState::draw(DrawerPtr& d)
{
	unsigned int i;
	TwoFloats shakeVal;
	ThreeFloats tempCol;
	const TwoFloats origin;

	if(!backgroundTweener.getFinished() && backgroundFlash)
		backgroundTweener.tick();
	d->fillRect(origin, d->getDrawingAreaSize(), backgroundColour);

	if(!blobRightTweener.getFinished())
		blobRightTweener.tick();
	else
	{
		ThreeFloats tempCol;

		while((tempCol == 0.0f) || (tempCol == 1.0f) || (tempCol == blobColourRight))
		{
			tempCol.x = (Random::getGlobal().getRandomFloat() > 0.5f) ? 1.0f : 0.0f;
			tempCol.y = (Random::getGlobal().getRandomFloat() > 0.5f) ? 1.0f : 0.0f;
			tempCol.z = (Random::getGlobal().getRandomFloat() > 0.5f) ? 1.0f : 0.0f;
		}
		blobRightTweener.reset(tempCol);
	}

	if(!blobLeftTweener.getFinished())
		blobLeftTweener.tick();
	else
		blobLeftTweener.reset(blobColourRight);

	if((shake > 0.0f) && screenShake)
	{
		//0.75f is the toned-down value.
		shakeVal.x = sinf(shakeIndex*2.0f*NIALL_PI) * 16.0f * shake * 0.75f;
		d->translate(shakeVal);
		d->rotate(shakeVal.x * 0.001f, TwoFloats(640.0f, 360.0f));

		shake -= 0.05f;
		shakeIndex += 0.1f;
		if(shakeIndex > 1.0f)
			shakeIndex -= 1.0f;
	}

	for(i=0;i<vaderExplosions.size();++i)
		vaderExplosions[i]->draw(d);

	for(i=0;i<blobVaders.size();++i)
		blobVaders[i]->draw(d, blobColourLeft, blobColourRight);

	for(i=0;i<vaderBullets.size();++i)
		vaderBullets[i]->draw(d);

	for(i=0;i<playerBullets.size();++i)
		playerBullets[i]->draw(d);

	tempCol.x = 1.0f-blobColourLeft.x;
	tempCol.y = 1.0f-blobColourLeft.y;
	tempCol.z = 1.0f-blobColourLeft.z;
	player->draw(d, tempCol);

	particles.draw(d);

	//Draw the score.
	if(0)
	{
		wstringstream tempstr;

		tempstr << L"Score: " << globalData->getGameCounter().getScore();

		d->drawText(tempstr.str(),
					L"Font0",
					TwoFloats(10.0f, 44.0f),
					384.0f,
					TextLayout::Left,
					1.0f,
					ThreeFloats(1.0f));
	}

	if((shake > 0.0f) && screenShake)
	{
		shakeVal.x = -shakeVal.x;
		d->translate(shakeVal);
		d->rotate(-shakeVal.x * 0.001f, TwoFloats(640.0f, 360.0f));
	}

	if(gameOver)
	{
		d->drawText(gameOverText,
					L"Font0",
					gameOverLayout,
					TwoFloats(0.0f, 320.0f),
					d->getDrawingAreaSize().x,
					TextLayout::Centre,
					1.0f,
					ThreeFloats(1.0f));
	}
	else if(gameWin)
	{
		d->drawText(gameWinText,
					L"Font0",
					gameWinLayout,
					TwoFloats(0.0f, 320.0f),
					d->getDrawingAreaSize().x,
					TextLayout::Centre,
					1.0f,
					ThreeFloats(1.0f));
	}
}

//------------------------------------------------------------------------------
void NewGamePlayState::keyDown(const wstring& action,
							const wstring& keyPressed,
							int keyModifier)
{
	if(action == L"Escape")
		nextState(L"TitleScreen");
}

//------------------------------------------------------------------------------
void NewGamePlayState::keyUp(const wstring& action,
							const wstring& keyPressed,
							int keyModifier)
{
	
}

//------------------------------------------------------------------------------
void NewGamePlayState::mouseDown(int button, const TwoFloats& pos)
{
	if(gameWin && !endPause)
	{
		globalData->incNumFleetsDefeated();
		globalData->incrementDifficulty();
		nextState(L"SolarSystemState");
		return;
	}
	else if(!gameOver)
	{
		firePlayerBullet();

		autofire = true;
		autofireCount = autofireMax;
	}
	else if(!endPause)
	{
		nextState(L"TitleScreen");
		return;
	}
}

//------------------------------------------------------------------------------
void NewGamePlayState::mouseUp(int button, const TwoFloats& pos)
{
	autofire = false;
}

//------------------------------------------------------------------------------
void NewGamePlayState::mouseMove(const TwoFloats& pos)
{
	player->mouseMove(pos);
}

//------------------------------------------------------------------------------
void NewGamePlayState::beatCallback()
{
	blobMove = true;
}

//------------------------------------------------------------------------------
void NewGamePlayState::countReached(const wstring& counter, int value)
{
	int i, j;
	int tempint;
	float tempf;
	int startVaders = 0;
	const TwoFloats origin;
	vector<int> availableFreqs;

	if(counter == L"Dead Vader")
	{
		if(musicAction > 0.33f)
		{
			for(i=0;i<8;++i)
				availableFreqs.push_back(i);

			for(i=0;i<4;++i)
			{
				tempf = 0.0f;
				for(j=startVaders;j<(startVaders+((int)blobVaders.size()/4));++j)
				{
					tempf += origin.getDistance(blobVaders[j]->getPosition());
				}
				startVaders += (blobVaders.size()/4);

				tempf *= 0.25f;
				tempf /= 1468.60478f;
				tempf *= ((float)availableFreqs.size())-0.01f;
				tempint = (int)tempf;
				if(tempint >= (int)availableFreqs.size())
					tempint = availableFreqs.size()-1;

				sounder->setPadStep(i, availableFreqs[tempint]);

				availableFreqs.erase(availableFreqs.begin()+tempint);
			}
		}
		else
		{
			for(i=0;i<4;++i)
				sounder->setPadStep(i, -1);
		}
	}
}

//------------------------------------------------------------------------------
void NewGamePlayState::addVader(const shared_ptr<Vader>& vader)
{
	blobVaders.push_back(vader);
	collidables.push_back(weak_ptr<Collidable>(blobVaders.back()));
}

//------------------------------------------------------------------------------
void NewGamePlayState::addVaderProjectile(const shared_ptr<VaderProjectile>& projectile)
{
	vaderBullets.push_back(projectile);
	collidables.push_back(weak_ptr<Collidable>(vaderBullets.back()));
	sounder->triggerSFX(1);
}

//------------------------------------------------------------------------------
void NewGamePlayState::addVaderExplosion(const shared_ptr<VaderExplosion>& explosion)
{
	vaderExplosions.push_back(explosion);
}

//------------------------------------------------------------------------------
void NewGamePlayState::collisionDetection()
{
	float dist;
	unsigned int i, j;
	std::vector<shared_ptr<Collidable> > tempCol;

	for(i=0;i<collidables.size();++i)
	{
		if(!collidables[i].expired())
			tempCol.push_back(shared_ptr<Collidable>(collidables[i]));
	}

	for(i=0;i<tempCol.size();++i)
	{
		shared_ptr<Collidable>& collidableI(tempCol[i]);

		if(collidableI->doNotCount())
			continue;

		for(j=(i+1);j<tempCol.size();++j)
		{
			shared_ptr<Collidable>& collidableJ(tempCol[j]);

			if(collidableJ->doNotCount())
				continue;

			dist = collidableI->getPosition().getDistance(collidableJ->getPosition());

			if(dist < (collidableI->getRadius() + collidableJ->getRadius()))
			{
				collidableI->collided(collidableJ->getName(),
									  collidableJ->getPosition());
				collidableJ->collided(collidableI->getName(),
									  collidableI->getPosition());
			}
		}
	}
}

//------------------------------------------------------------------------------
void NewGamePlayState::addNewVaders()
{
	int i;
	TwoFloats tempPos;

	if(remainingWaves > 0)
	{
		if(waveTypes.size() > 0)
		{
			for(i=1;i<8;++i)
			{
				tempPos.x = 128.0f+((getDrawingAreaSize().x-256.0f) * ((float)i/8.0f));
				tempPos.y = (72.0f+((640-256.0f) * ((float)0.0f/4.0f))) - 448.0f;

				blobVaders.push_back(globalData->getVaderFactory().createVader(waveTypes[0],
																			   tempPos,
																			   *this));
				collidables.push_back(weak_ptr<Vader>(blobVaders.back()));
			}
			waveTypes.erase(waveTypes.begin());
		}

		--remainingWaves;
	}
}

//------------------------------------------------------------------------------
void NewGamePlayState::addNote(const TwoFloats& pos)
{
	int note;
	int cell;
	int synth = 0;
	const TwoFloats origin;
	const float maxDist = 1468.60478f;

	cell = (notePos + (int)killDistance + (int)((1.0f-musicAction) * 4.0f)) % 16;
	note = (int)((origin.getDistance(pos)/maxDist) * 15.99f);
	if(note > 7)
	{
		synth = 1;
		note -= 8;
	}

	sounder->addNote(synth, cell, note);
	if(synth && (Random::getGlobal().getRandomFloat() > 0.5f))
	{
		cell += 8;
		cell %= 16;
		sounder->addNote(synth, cell, note);
	}

	notePos = (cell+1)%16;
	killDistance = 0.0f;
}

//------------------------------------------------------------------------------
void NewGamePlayState::firePlayerBullet()
{
	TwoFloats dir;
	ThreeFloats tempCol;
	float speed = player->getSpeed();

	dir.x = sinf(speed * 0.2f);
	dir.y = -cosf(speed * 0.2f);

	playerBullets.push_back(shared_ptr<PlayerBullet>(new PlayerBullet(player->getPosition(), dir)));
	collidables.push_back(weak_ptr<Collidable>(playerBullets.back()));
	player->fired();

	sounder->triggerSFX(0);
}
