//	NewGamePlayState.h - The main gameplay state.
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2012  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef NEWGAMEPLAYSTATE_H_
#define NEWGAMEPLAYSTATE_H_

#include "drawing/TextLayout.h"
#include "ParticleGenerator.h"
#include "BlobVaderSounder.h"
#include "ThreeFloats.h"
#include "GameCounter.h"
#include "GameState.h"
#include "Tweener.h"
#include "Player.h"

#include <stdint.h>
#include <vector>

class VaderProjectile;
class VaderExplosion;
class BlobGlobalData;
class PlayerBullet;
class Vader;

///	The main gameplay state.
class NewGamePlayState : public GameState,
						 public CounterNotification,
						 public BlobVaderSounder::Callback
{
  public:
	///	Constructor.
	NewGamePlayState(Application& app, SounderPtr& sound);
	///	Destructor.
	~NewGamePlayState();

	///	Called frequently to update the state.
	void update();
	///	Called to draw the state.
	void draw(DrawerPtr& d);

	///	Called when a key down event is received.
	void keyDown(const std::wstring& action,
				 const std::wstring& keyPressed,
				 int keyModifier);
	///	Called when a key up event is received.
	void keyUp(const std::wstring& action,
			   const std::wstring& keyPressed,
			   int keyModifier);
	

	///	Called when a mouse down event is received.
	void mouseDown(int button, const TwoFloats& pos);
	///	Called when a mouse down event is received.
	void mouseUp(int button, const TwoFloats& pos);
	///	Used to test connected line drawing.
	void mouseMove(const TwoFloats& pos);

	///	Called when it's time to advance the BlobVaders.
	void beatCallback();
	///	Called when the related counter reaches the specified value.
	void countReached(const std::wstring& counter, int value);

	///	Adds a new vader to the state.
	void addVader(const shared_ptr<Vader>& vader);
	///	Called by Vaders when they shoot a VaderProjectile.
	void addVaderProjectile(const shared_ptr<VaderProjectile>& projectile);
	///	Adds a VaderExplosion.
	void addVaderExplosion(const shared_ptr<VaderExplosion>& explosion);

	///	Returns the length of time elapsed since the state's creation.
	float getTimeElapsed() const {return timeElapsed;};

	///	Returns the name of this state.
	const std::wstring getName() const {return L"NewGamePlayState";};
	///	Used to construct instances of this state.
	static GameStateCreator<NewGamePlayState> creator;
  private:
	///	Helper method. Does all the collision detection for the game.
	void collisionDetection();
	///	Helper method. Adds a new row of BlobVaders.
	void addNewVaders();
	///	Helper method. Adds a note based on a position on screen.
	void addNote(const TwoFloats& pos);
	///	Helper method. Fires a player bullet.
	void firePlayerBullet();

	///	The current background colour.
	ThreeFloats backgroundColour;
	///	Tweener for the background colour.
	STweener<ThreeFloats> backgroundTweener;

	///	The BlobVaders.
	std::vector<shared_ptr<Vader> > blobVaders;
	///	The BlobVaders left colour.
	ThreeFloats blobColourLeft;
	///	Tweener for the BlobVaders left colour.
	STweener<ThreeFloats> blobLeftTweener;
	///	The BlobVaders right colour.
	ThreeFloats blobColourRight;
	///	Tweener for the BlobVaders right colour.
	STweener<ThreeFloats> blobRightTweener;
	///	Used to move the BlobVaders.
	int blobMoveIndex;
	///	Used to move the BlobVaders.
	int blobMoveCount;
	///	Tells us to move the BlobVaders in the next update() call.
	bool blobMove;
	///	Any active VaderBullet objects.
	std::vector<shared_ptr<VaderProjectile> > vaderBullets;
	///	Used to determine when to add new VaderBullets.
	int vaderBulletCount;
	///	Used to determine when to add new VaderBullets.
	int vaderBulletInc;
	///	Used to speed up the time between vaders firing.
	float vaderBulletSpeed;

	///	Any explosions from dead Vaders.
	std::vector<shared_ptr<VaderExplosion> > vaderExplosions;

	///	Used to determine how many waves of vaders are in this fleet.
	int remainingWaves;
	///	The type of each of the vader waves.
	std::vector<std::wstring> waveTypes;

	///	The player object.
	shared_ptr<Player> player;
	///	Any active PlayerBullet objects.
	std::vector<shared_ptr<PlayerBullet> > playerBullets;

	///	Used for collision detection.
	std::vector<weak_ptr<Collidable> > collidables;

	///	Used to shake the screen.
	float shake;
	///	Used to shake the screen.
	float shakeIndex;
	///	The particle generator.
	ParticleGenerator particles;

	///	True if the player's dead.
	bool gameOver;
	///	Game over text.
	std::wstring gameOverText;
	///	Game over TextLayout.
	TextLayout gameOverLayout;

	///	True if the player defeated the fleet.
	bool gameWin;
	///	Game win text.
	std::wstring gameWinText;
	///	Game win TextLayout.
	TextLayout gameWinLayout;

	///	Used to make the music more interesting as the game gets more hectic.
	float musicAction;
	///	Used to determine where to add notes.
	int notePos;
	///	Used to determine where to add notes.
	float killDistance;

	///	The length of time elapsed since the state was created.
	float timeElapsed;
	///	The previous ticks value.
	uint32_t lastTicks;

	///	Autofire variable.
	bool autofire;
	///	Autofire variable.
	int autofireCount;
	///	Autofire variable.
	int autofireMax;

	///	Our sounder.
	shared_ptr<BlobVaderSounder> sounder;
	///	The global data for the game.
	shared_ptr<BlobGlobalData> globalData;
	///	True if the background can flash.
	bool backgroundFlash;
	///	True if the screen can shake.
	bool screenShake;

	///	Used to make sure the player doesn't restart by accident.
	int endPause;
};

#endif
