//	Planet.cpp - A planet in the solar system.
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2012  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "drawing/Drawer.h"
#include "MiscHelper.h"
#include "Planet.h"
#include "Random.h"

//------------------------------------------------------------------------------
Planet::Planet(float rad, const TwoFloats& sunPos, int sine):
radius(rad),
colour(1.0f, 0.0f, 0.0f),
health(1.0f),
sineIndex(sine),
scale(1.0f),
scaleTweener(scale, scale, 0.05f),
colourTweener(colour, ThreeFloats(0.0f, 0.0f, 1.0f), 0.01f)
{
	ThreeFloats tempCol;
	float index = Random::getGlobal().getRandomFloat();

	position = sunPos;
	position.x += sinf(index * 2.0f * NIALL_PI) * radius;
	position.y += cosf(index * 2.0f * NIALL_PI) * radius;

	size = 0.8f + (Random::getGlobal().getRandomFloat() * 0.5f);

	while((tempCol == 0.0f) || (tempCol == 1.0f))
	{
		tempCol.x = (Random::getGlobal().getRandomFloat() > 0.5f) ? 1.0f : 0.0f;
		tempCol.y = (Random::getGlobal().getRandomFloat() > 0.5f) ? 1.0f : 0.0f;
		tempCol.z = (Random::getGlobal().getRandomFloat() > 0.5f) ? 1.0f : 0.0f;
	}
	if(tempCol.x < 0.5f)
		tempCol.x += 0.25f;
	if(tempCol.y < 0.5f)
		tempCol.y += 0.25f;
	if(tempCol.z < 0.5f)
		tempCol.z += 0.25f;
	colour = tempCol;

	while((tempCol == 0.0f) || (tempCol == 1.0f) || (tempCol == colour))
	{
		tempCol.x = (Random::getGlobal().getRandomFloat() > 0.5f) ? 1.0f : 0.0f;
		tempCol.y = (Random::getGlobal().getRandomFloat() > 0.5f) ? 1.0f : 0.0f;
		tempCol.z = (Random::getGlobal().getRandomFloat() > 0.5f) ? 1.0f : 0.0f;
	}
	if(tempCol.x < 0.5f)
		tempCol.x += 0.33f;
	if(tempCol.y < 0.5f)
		tempCol.y += 0.33f;
	if(tempCol.z < 0.5f)
		tempCol.z += 0.33f;
	colourTweener.reset(tempCol);

	regenerateName();
}

//------------------------------------------------------------------------------
Planet::~Planet()
{

}

//------------------------------------------------------------------------------
void Planet::draw(shared_ptr<Drawer>& d, const TwoFloats& screenPos)
{
	const TwoFloats origin(-32.0f);
	const TwoFloats centre(d->getDrawingAreaSize()*0.5f);
	TwoFloats tempPos(position-(16.0f*size*scale)-screenPos);
	//const float angle = position.getAngle(centre);
	float angle;
	TwoFloats temp1, temp2;

	temp1.x = position.x;
	temp1.y = centre.y;
	temp2.x = centre.x;
	temp2.y = position.y;
	angle = temp2.getAngle(temp1);

	if(tempPos.isWithin(origin, d->getDrawingAreaSize()+64.0f))
	{
		if(!colourTweener.getFinished())
			colourTweener.tick();
		else
		{
			ThreeFloats tempCol;

			while((tempCol == 0.0f) || (tempCol == 1.0f) || (tempCol == colour))
			{
				tempCol.x = (Random::getGlobal().getRandomFloat() > 0.5f) ? 1.0f : 0.0f;
				tempCol.y = (Random::getGlobal().getRandomFloat() > 0.5f) ? 1.0f : 0.0f;
				tempCol.z = (Random::getGlobal().getRandomFloat() > 0.5f) ? 1.0f : 0.0f;
			}
			if(tempCol.x < 0.5f)
				tempCol.x += 0.33f;
			if(tempCol.y < 0.5f)
				tempCol.y += 0.33f;
			if(tempCol.z < 0.5f)
				tempCol.z += 0.33f;
			colourTweener.reset(tempCol);
		}

		if(!scaleTweener.getFinished())
			scaleTweener.tick();

		d->drawImage(L"Blob", tempPos, TwoFloats(size*scale), colour);
		d->drawRotImage(L"BlobShadow",
						tempPos,
						angle,
						TwoFloats(32.0f),
						TwoFloats(size*scale*0.5f),
						ThreeFloats(0.0f),
						0.5f);
		d->drawRotImage(L"BlobShadow",
						tempPos,
						angle+NIALL_PI,
						TwoFloats(32.0f),
						TwoFloats(size*scale*0.5f),
						ThreeFloats(1.0f, 1.0f, 0.0f),
						0.33f);

		if(scale > 1.0f)
		{
			tempPos = position-screenPos;
			tempPos.x -= 128.0f;
			tempPos.y += (16.0f*size*scale) + 20.0f;

			d->drawText(name,
						L"Font2",
						nameLayout,
						tempPos,
						256.0f,
						TextLayout::Centre,
						1.0f,
						ThreeFloats(0.8f),
						(scale-1.0f)*2.0f);
		}
	}
}

//------------------------------------------------------------------------------
bool Planet::incrementHealth(float val)
{
	bool retval = false;

	health += val;
	if(health <= 0.0f)
		retval = true;

	return retval;
}

//------------------------------------------------------------------------------
void Planet::mouseOver(bool val)
{
	if(val)
	{
		if((scale < 1.5f) && (scaleTweener.getEndVal() < 1.5f))
		{
			scaleTweener.setDelta(0.05f);
			scaleTweener.reset(1.5f);
		}
	}
	else
	{
		if((scale > 1.0f) && (scaleTweener.getEndVal() > 1.0f))
		{
			scaleTweener.setDelta(0.02f);
			scaleTweener.reset(1.0f);
		}
	}
}

//------------------------------------------------------------------------------
void Planet::regenerateName()
{
	int index;
	const std::wstring first[] = {L"Sa",
								  L"Ur",
								  L"Ti",
								  L"Ea",
								  L"Ma",
								  L"Ne",
								  L"So",
								  L"Ve",
								  L"Plu",
								  L"Su",
								  L"Me",
								  L"Po",
								  L"Tu",
								  L"Ye",
								  L"Ka",
								  L"La"};
	const std::wstring second[] = {L"tu",
								   L"ro",
								   L"ri",
								   L"ptu",
								   L"tu",
								   L"rt",
								   L"ra",
								   L"ta",
								   L"nu",
								   L"lla",
								   L"yi",
								   L"re",
								   L"ce",
								   L"ze",
								   L"ho",
								   L"rta"};
	const std::wstring third[] = {L"rn",
								  L"to",
								  L"nus",
								  L"h",
								  L"l",
								  L"n",
								  L"zin",
								  L"sol",
								  L"this",
								  L"world",
								  L"tim",
								  L"a",
								  L"ch",
								  L"ter",
								  L"roid",
								  L"ah"};

	index = Random::getGlobal().getRandomInt(16);
	name = first[index];
	index = Random::getGlobal().getRandomInt(16);
	name.append(second[index]);
	index = Random::getGlobal().getRandomInt(16);
	name.append(third[index]);
}
