//	Planet.h - A planet in the solar system.
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2012  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef PLANET_H_
#define PLANET_H_

#include "drawing/TextLayout.h"
#include "Array2DDrawer.h"
#include "ThreeFloats.h"
#include "c0xHeaders.h"
#include "TwoFloats.h"
#include "Tweener.h"

#include <string>

class Drawer;

///	A planet in the solar system.
class Planet
{
  public:
	///	Constructor.
	/*!
		\param rad The planet's radius out from the sun.
		\param sunPos The position of the sun (in pixels). Used to determine the
		planet's position.
		\param sine The planet's index to it's corresponding BeatingSine.

		The planet's position will be a random position along the circle with
		this radius, set in the constructor.
	 */
	Planet(float rad, const TwoFloats& sunPos, int sine);
	///	Destructor.
	~Planet();

	///	Draws the planet.
	void draw(shared_ptr<Drawer>& d, const TwoFloats& screenPos);

	///	Increments the planet's health.
	/*!
		\param val The amount to increment the planet's health. A negative
		value will decrement it.
		\return Will return true if the planet has lost all its health and
		must be destroyed.
	 */
	bool incrementHealth(float val);

	///	Tells the planet whether the mouse is over it.
	void mouseOver(bool val);

	///	Called to generate a new name for the planet.
	void regenerateName();

	///	Returns the planet's position.
	const TwoFloats& getPosition() const {return position;};
	///	Returns the planet's radius out from the sun.
	float getRadius() const {return radius;};
	///	Returns the planet's colour.
	const ThreeFloats& getColour() const {return colour;};
	///	Returns the planet's health.
	float getHealth() const {return health;};
	///	Returns the planet's name.
	const std::wstring& getName() const {return name;};
	///	Returns the planet's sine index.
	int getSineIndex() const {return sineIndex;};
	///	Returns the planet's scale.
	float getScale() const {return scale;};
  private:
	///	The planet's position.
	TwoFloats position;
	///	The planet's radius out from the sun.
	float radius;
	///	The planet's size.
	float size;
	///	The planet's colour.
	ThreeFloats colour;
	///	The planet's health.
	float health;
	///	The planet's name.
	std::wstring name;
	///	The index to the planet's corresponding BeatingSine.
	int sineIndex;

	///	The planet's scale (for when the mouse is over it).
	float scale;
	///	The planet's scale tweener.
	STweener<float> scaleTweener;

	///	The planet's colour tweener.
	STweener<ThreeFloats> colourTweener;

	///	Used to draw the planet's name.
	TextLayout nameLayout;
};

#endif
