//	SolarSystemState.cpp - The solar system overview state.
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2012  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "infoBoxes/EncyclopediaBox.h"
#include "infoBoxes/PlanetBox.h"
#include "SolarSystemState.h"
#include "BlobVaderSounder.h"
#include "drawing/Drawer.h"
#include "BlobGlobalData.h"
#include "Application.h"
#include "MiscHelper.h"
#include "Planet.h"
#include "Random.h"
#include "Fleet.h"

#include <iomanip>
#include <cassert>
#include <sstream>

using std::wstringstream;

//Temp.
namespace PauseButtonImage
{
const int width = 7;
const int height = 7;
const char pixels[] = {	0,0,0,0,0,0,0,
						0,1,1,0,1,1,0,
						0,1,1,0,1,1,0,
						0,1,1,0,1,1,0,
						0,1,1,0,1,1,0,
						0,1,1,0,1,1,0,
						0,0,0,0,0,0,0};
};

//------------------------------------------------------------------------------
GameStateCreator<SolarSystemState> SolarSystemState::creator;

//------------------------------------------------------------------------------
SolarSystemState::SolarSystemState(Application& app, SounderPtr& sound):
GameState(app, sound),
backgroundColour(0.1f),
backgroundTweener(backgroundColour, ThreeFloats(0.1f), 0.05f),
shake(0.0f),
shakeIndex(0.0f),
daysElapsed(0.0f),
timePaused(true),
mousePos(640.0f, 360.0f),
displayInfo(true),
infoSize(422.0f, 523.0f),
currentInfoBox(PlanetBox),
globalData(dynamic_pointer_cast<BlobGlobalData>(app.getGlobalData())),
gameOver(false),
sounder(dynamic_pointer_cast<BlobVaderSounder>(sound))
{
	float dist;
	int tempint;
	unsigned int i, j;
	float radius = 0.2f;
	float minVal, maxVal;
	const float minDistance = 384.0f;
	const float maxDistance = 1280.0f;
	const TwoFloats sunPos = getDrawingAreaSize() * 0.5f;

	infoPos.x = (getDrawingAreaSize().x*0.5f) - 211.0f;
	infoPos.y = (getDrawingAreaSize().y*0.5f) - 261.5f;
	infoClosePos.x = infoPos.x+infoSize.x - (12.0f + 31.0f);
	infoClosePos.y = infoPos.y + 12.0f;

	infoBoxes[EncyclopediaBox] = shared_ptr<InfoBox>(new EncyclopediaInfoBox(globalData,
																			 getDrawingAreaSize()));

	globalData = dynamic_pointer_cast<BlobGlobalData>(getGlobalData());
	setMousePos(mousePos);

	screenShake = globalData->getScreenShake();
	backgroundFlash = globalData->getFlashingBackground();

	sounder->activateBrushStrokes(true);
	for(i=0;i<BrushStrokesSynth::TableSize;++i)
		mouseTrail[i] = mousePos;

	if(globalData->getPlanets().size() == 0)
	{
		//Add the planets.
		for(i=0;i<(NumPlanets-1);++i)
		{
			planets.push_back(shared_ptr<Planet>(new Planet(radius*280.0f,
												 getDrawingAreaSize()*0.5f,
												 i)));
			planetOffsets.push_back(planets.back()->getRadius()/320.0f * 0.1f);

			minVal = radius + 0.1f;
			maxVal = minVal + ((1.0f-minVal)/((float)(NumPlanets-1-i)));
			radius = minVal + (Random::getGlobal().getRandomFloat() * (maxVal-minVal));
		}
		radius = 1.0f;
		planets.push_back(shared_ptr<Planet>(new Planet(radius*280.0f,
												 getDrawingAreaSize()*0.5f,
												 i)));
		planetOffsets.push_back(planets.back()->getRadius()/320.0f * 0.1f);

		//Add the fleets.
		for(i=0;i<planets.size();++i)
		{
			dist = (1.0f-((float)i/(float)planets.size())) * (maxDistance-minDistance);
			dist += minDistance;
			fleets.push_back(shared_ptr<Fleet>(new Fleet(planets[i]->getPosition(),
											   sunPos,
											   dist)));
			fleets.back()->setSynth(i);
		}

		player.setInitialPosition(planets[0]->getPosition());
		playerTrajectory = planets[0]->getPosition();
	}
	else
	{
		for(i=0;i<globalData->getPlanets().size();++i)
		{
			planets.push_back(globalData->getPlanets()[i]);
			planetOffsets.push_back(planets.back()->getRadius()/320.0f * 0.1f);
		}
		for(i=0;i<globalData->getFleets().size();++i)
			fleets.push_back(globalData->getFleets()[i]);
		player.setInitialPosition(globalData->getPlayerPosition());
		playerTrajectory = player.getPosition();
	}

	infoBoxes[PlanetBox] = shared_ptr<InfoBox>(new PlanetInfoBox(planets[0],
																 getDrawingAreaSize()));
	tempint = getPlayerPlanet();
	if(tempint > -1)
		dynamic_pointer_cast<PlanetInfoBox>(infoBoxes[PlanetBox])->setPlanet(planets[tempint]);
	else
		displayInfo = false;

	//Setup the buttons.
	for(i=0;i<5;++i)
	{
		for(j=0;j<5;++j)
		{
			playButton[i][j] = false;
			pauseButton[i][j] = false;
			encyclopediaButton[i][j] = false;
		}
	}

	playButton[0][1] = true;
	playButton[1][1] = true;
	playButton[2][1] = true;
	playButton[3][1] = true;
	playButton[4][1] = true;
	playButton[1][2] = true;
	playButton[2][2] = true;
	playButton[3][2] = true;
	playButton[2][3] = true;

	pauseButton[0][1] = true;
	pauseButton[1][1] = true;
	pauseButton[2][1] = true;
	pauseButton[3][1] = true;
	pauseButton[4][1] = true;
	pauseButton[0][3] = true;
	pauseButton[1][3] = true;
	pauseButton[2][3] = true;
	pauseButton[3][3] = true;
	pauseButton[4][3] = true;

	encyclopediaButton[0][1] = true;
	encyclopediaButton[1][1] = true;
	encyclopediaButton[2][1] = true;
	encyclopediaButton[3][1] = true;
	encyclopediaButton[4][1] = true;
	encyclopediaButton[0][2] = true;
	encyclopediaButton[2][2] = true;
	encyclopediaButton[4][2] = true;
	encyclopediaButton[0][3] = true;
	encyclopediaButton[4][3] = true;
}

//------------------------------------------------------------------------------
SolarSystemState::~SolarSystemState()
{
	int i;

	for(i=0;i<6;++i)
	{
		sounder->setSineAmplitude(i, 0.0f);
		sounder->activateFleet(i, false);
	}

	sounder->activateBrushStrokes(false);
	sounder->activateSun(false);
}

//------------------------------------------------------------------------------
void SolarSystemState::update()
{
	int tempint;
	unsigned int i;
	bool lastPaused = timePaused;
	const TwoFloats pos = mousePos - screenPos;
	const TwoFloats drawSize = getDrawingAreaSize();

	if(player.getShipIsMoving() && !gameOver)
		timePaused = player.update();

	//Update the mouse trail for the brush strokes synth.
	for(i=(BrushStrokesSynth::TableSize-1);i>0;--i)
		mouseTrail[i] = mouseTrail[i-1];
	if(mouseTrail[0] != pos)
	{
		mouseTrail[0] += (pos-mouseTrail[0]) * 0.1f;
		if((pos-mouseTrail[0]).abs() < 0.1f)
			mouseTrail[0] = pos;
	}
	else
		mouseTrail[0] = pos;
	sounder->setBrushStrokesTrail(mouseTrail);

	if(!lastPaused && timePaused)
	{
		tempint = getPlayerPlanet();
		if(tempint > -1)
		{
			displayInfo = true;

			currentInfoBox = PlanetBox;
			dynamic_pointer_cast<PlanetInfoBox>(infoBoxes[PlanetBox])->setPlanet(planets[tempint]);
		}
	}

	if(!timePaused)
	{
		for(i=0;i<fleets.size();++i)
			fleets[i]->update();

		//Check whether any fleets have collided with the player.
		for(i=0;i<fleets.size();++i)
		{
			if(fleets[i]->getPosition().getDistance(player.getPosition()) < 48.0f)
			{
				globalData->storePlanets(planets);
				globalData->storeFleets(fleets);
				globalData->setCurrentFleetIndex(i);
				globalData->setPlayerPosition(player.getPosition());

				nextState(L"NewGamePlayState");
				return;
			}
		}

		//Check whether any fleets have collided with any planets.
		planetFleetCollision();

		//It takes 5 days to cover the diameter of the solar system.
		//Each frame counts as 0.001 of the distance travelled.
		//Therefore each frame is (0.001*5) days: 0.005.
		daysElapsed += 0.005f;
	}

	//Update particles.
	particles.update();

	//Move screen if the mouse is at the edge.
	if((pos.x < 16.0f) && (screenPos.x > -1280.0f))
		screenPos.x -= 4.0f;
	else if((pos.x > (drawSize.x-16.0f)) && (screenPos.x < (drawSize.x + 1280.0f)))
		screenPos.x += 4.0f;
	if((pos.y < 16.0f) && (screenPos.y > -1280.0f))
		screenPos.y -= 4.0f;
	else if((pos.y > (drawSize.y-16.0f)) && (screenPos.y < (drawSize.y + 1280.0f)))
		screenPos.y += 4.0f;

	mousePos = screenPos + pos;

	//Mouse over any planets.
	if(displayInfo)
		tempint = -1;
	else
		tempint = getPlanetUnderMouse();
	for(i=0;i<planets.size();++i)
	{
		planets[i]->mouseOver(tempint == i);
		planetOffsets[i] += (planets[i]->getScale()-1.0f) * 0.001f;

		if(planets[i]->getScale() > 1.0f)
		{
			sounder->setSineAmplitude(planets[i]->getSineIndex(),
									  (planets[i]->getScale()-1.0f) * 0.25f);
		}
	}

	//Mouse over the sun.
	if(mousePos.getDistance(getDrawingAreaSize()*0.5f) < 32.0f)
	{
		if(!displayInfo)
		{
			sun.mouseOver(true);
			sounder->activateSun(true);
		}
	}
	else
	{
		sun.mouseOver(false);
		sounder->activateSun(false);
	}

	//Mouse over any fleets.
	for(i=0;i<fleets.size();++i)
	{
		if(mousePos.getDistance(fleets[i]->getPosition()) < 32.0f)
		{
			sounder->activateFleet(fleets[i]->getSynth(), true);
			fleets[i]->mouseOver(true);
		}
		else
		{
			sounder->activateFleet(fleets[i]->getSynth(), false);
			fleets[i]->mouseOver(false);
		}

		if(displayInfo)
		{
			sounder->activateFleet(fleets[i]->getSynth(), false);
			fleets[i]->mouseOver(false);
		}
	}

	if(displayInfo)
	{
		sun.mouseOver(false);
		sounder->activateSun(false);
	}
}

//------------------------------------------------------------------------------
void SolarSystemState::draw(DrawerPtr& d)
{
	int j;
	float rad;
	//float offset;
	int numPoints;
	unsigned int i;
	TwoFloats tempPos;
	TwoFloats lastPos;
	TwoFloats shakeVal;
	ThreeFloats tempCol;
	wstringstream tempstr;
	const TwoFloats origin;
	const int maxPoints = 128;
	const TwoFloats centre = d->getDrawingAreaSize() * 0.5f;
	const TwoFloats encyclopediaPos(d->getDrawingAreaSize().x-32.0f, 64.0f);

	//Draw background.
	if(!backgroundTweener.getFinished() && backgroundFlash)
		backgroundTweener.tick();
	d->fillRect(origin, d->getDrawingAreaSize(), backgroundColour);

	//Handle any screen shake.
	if((shake > 0.0f) && screenShake)
	{
		//0.75f is the toned-down value.
		shakeVal.x = sinf(shakeIndex*2.0f*NIALL_PI) * 16.0f * shake * 0.75f;
		d->translate(shakeVal);
		d->rotate(shakeVal.x * 0.001f, TwoFloats(640.0f, 360.0f));

		shake -= 0.05f;
		shakeIndex += 0.1f;
		if(shakeIndex > 1.0f)
			shakeIndex -= 1.0f;
	}

	//Draw planet trajectories.
	for(i=0;i<planets.size();++i)
	{
		rad = planets[i]->getRadius();
		//offset = rad/320.0f * 0.1f;

		numPoints = 8 + (int)((float)(maxPoints-8) * rad/320.0f);
		for(j=0;j<numPoints;j+=2)
		{
			tempPos = centre-screenPos;
			tempPos.x += sinf((((float)j/(float)numPoints)+planetOffsets[i]) * 2.0f * NIALL_PI) * rad;
			tempPos.y += -cosf((((float)j/(float)numPoints)+planetOffsets[i]) * 2.0f * NIALL_PI) * rad;
			d->drawImage(L"Blob",
						 tempPos-(16.0f*0.33f),
						 TwoFloats(0.33f),
						 planets[i]->getColour(),
						 0.33f);
		}
	}

#ifndef DEBUG
	//Draw sun.
	sun.draw(d, screenPos);
#endif

	//Draw player trajectory.
	if(playerTrajectory != player.getPosition())
	{
		playerTrajectoryLine.draw(d,
								  player.getPosition(),
								  playerTrajectory,
								  player.getColour(),
								  screenPos,
								  ((5.0f/560.0f)*(player.getPosition().getDistance(playerTrajectory))));
	}

	//If the player's mousing over a fleet, draw it's estimated trajectories.
	for(i=0;i<fleets.size();++i)
	{
		if(mousePos.getDistance(fleets[i]->getPosition()) < 32.0f)
		{
			int index1 = -1;
			int index2 = -1;
			TwoFloats planetPos;
			float dist1 = 10000.0f;
			float dist2 = 10000.0f;
			const TwoFloats fleetPos = fleets[i]->getPosition();

			for(j=0;j<(int)planets.size();++j)
			{
				planetPos = planets[j]->getPosition();
				if(fleetPos.getDistance(planetPos) < dist1)
				{
					index2 = index1;
					dist2 = dist1;

					index1 = j;
					dist1 = fleetPos.getDistance(planetPos);
				}
				else if(fleetPos.getDistance(planetPos) < dist2)
				{
					index2 = j;
					dist2 = fleetPos.getDistance(planetPos);
				}
			}

			if(index1 > -1)
			{
				fleetTrajectoryLines[0].draw(d,
											 fleetPos,
											 planets[index1]->getPosition(),
											 fleets[i]->getColour(),
											 screenPos,
											 ((5.0f/560.0f)*dist1));
			}
			if(index2 > -1)
			{
				fleetTrajectoryLines[0].draw(d,
											 fleetPos,
											 planets[index2]->getPosition(),
											 fleets[i]->getColour(),
											 screenPos,
											 ((5.0f/560.0f)*dist2));
			}

			break;
		}
	}

	//Draw planets.
	for(i=0;i<planets.size();++i)
		planets[i]->draw(d, screenPos);

	//Draw player's ship.
	player.draw(d, screenPos);

	//Draw fleets.
	for(i=0;i<fleets.size();++i)
		fleets[i]->draw(d, screenPos);

	//Draw particles.
	particles.draw(d);

	//Draw play/pause button.
	if(timePaused)
	{
		if((mousePos-screenPos) < 128.0f)
			tempCol = ThreeFloats(1.0f);
		else
			tempCol = ThreeFloats(0.75f);

		arrayDrawer.draw(d, playButton, 0.625f, TwoFloats(96.0f), tempCol);
	}
	else
	{
		if((mousePos-screenPos) < 128.0f)
			tempCol = ThreeFloats(1.0f);
		else
			tempCol = ThreeFloats(0.75f);

		arrayDrawer.draw(d, pauseButton, 0.625f, TwoFloats(96.0f), tempCol);
	}

	//Draw the encyclopedia button.
	tempPos = mousePos-screenPos;
	if((tempPos.x > (encyclopediaPos.x-80.0f)) && (tempPos.y < (encyclopediaPos.y+32.0f)))
		tempCol = ThreeFloats(1.0f);
	else
		tempCol = ThreeFloats(0.75f);
	arrayDrawer.draw(d, encyclopediaButton, 0.5f, encyclopediaPos, tempCol);

	//Draw the info box if necessary.
	if(displayInfo && (globalData->getNumFleetsDefeated() < 6))
	{
		d->drawImage(L"InfoBox", infoPos);
		d->drawImage(L"TitleX", infoClosePos);
		infoBoxes[currentInfoBox]->draw(d);
	}

	//Draw mouse.
	tempPos = mousePos - screenPos;
	d->drawImage(L"Blob", tempPos, TwoFloats(0.5f));
	tempPos.x += 10.0f;
	tempPos.y += 5.0f;
	d->drawImage(L"Blob", tempPos, TwoFloats(0.5f));
	tempPos.x -= 10.0f;
	tempPos.y += 5.0f;
	d->drawImage(L"Blob", tempPos, TwoFloats(0.5f));
	tempPos.y += 10.0f;
	d->drawImage(L"Blob", tempPos, TwoFloats(0.5f));
	tempPos.x += 18.0f;
	tempPos.y -= 11.0f;
	d->drawImage(L"Blob", tempPos, TwoFloats(0.5f));

	//Draw days elapsed.
	tempstr.clear();
	tempstr.str(L"");
	tempstr.precision(1);
	tempstr << L"Days Elapsed: " << std::fixed << daysElapsed;
	d->drawText(tempstr.str(),
				L"Font1",
				TwoFloats(16.0f, d->getDrawingAreaSize().y-20.0f),
				(d->getDrawingAreaSize().x-32.0f),
				TextLayout::Left,
				1.0f,
				ThreeFloats(0.75f));

	//Draw game over text.
	if(gameOver)
	{
		d->drawText(L"Game Over\n\nAll planets lost.\n\n\n\n\n<click to title>",
					L"Font0",
					gameOverLayout,
					TwoFloats(0.0f, 320.0f),
					d->getDrawingAreaSize().x,
					TextLayout::Centre,
					1.0f,
					ThreeFloats(1.0f));
	}
	else if(globalData->getNumFleetsDefeated() > 5)
	{
		d->drawText(L"Success!\n\nInvasion defeated.\n\n\n\n\n<click to title>",
					L"Font0",
					gameWinLayout,
					TwoFloats(0.0f, 320.0f),
					d->getDrawingAreaSize().x,
					TextLayout::Centre,
					1.0f,
					ThreeFloats(1.0f));
	}
}

//------------------------------------------------------------------------------
void SolarSystemState::keyDown(const std::wstring& action,
							   const std::wstring& keyPressed,
							   int keyModifier)
{
	if(displayInfo)
		displayInfo = false;
	else if(action == L"Escape")
		nextState(L"TitleScreen");
	else if(action == L"Action")
	{
		timePaused = !timePaused;
		player.setPause(timePaused);
		sounder->triggerSFX(0);
	}
}

//------------------------------------------------------------------------------
void SolarSystemState::mouseDown(int button, const TwoFloats& pos)
{
	int mousePlanet;
	const TwoFloats encyclopediaPos(getDrawingAreaSize().x-128.0f, 128.0f);

	mousePos = screenPos + pos;

	if(!gameOver && (globalData->getNumFleetsDefeated() < 6))
	{
		if((pos.x > encyclopediaPos.x) && (pos.y < encyclopediaPos.y))
		{
			currentInfoBox = EncyclopediaBox;
			displayInfo = true;
			sounder->triggerSFX(0);
		}
		else if(!displayInfo)
		{
			mousePlanet = getPlanetUnderMouse();
			if(mousePlanet > -1)
			{
				player.setDestination(planets[mousePlanet]->getPosition());
				playerTrajectory = planets[mousePlanet]->getPosition();
				timePaused = false;
				player.setPause(timePaused);

				if(backgroundFlash)
				{
					backgroundColour = planets[mousePlanet]->getColour() * 0.25f;
					backgroundTweener.reset(ThreeFloats(0.1f));
				}
			}
			else if(pos < 128.0f)
			{
				timePaused = !timePaused;
				player.setPause(timePaused);
				sounder->triggerSFX(0);
			}
		}
		else if((pos > infoPos) && (pos < (infoPos+infoSize)))
		{
			if((pos > infoClosePos) && (pos < (infoClosePos+31.0f)))
			{
				displayInfo = false;
				sounder->triggerSFX(0);
			}
			else
				infoBoxes[currentInfoBox]->mouseDown(button, pos);
		}
		else
			displayInfo = false;
	}
	else
		nextState(L"TitleScreen");
}

//------------------------------------------------------------------------------
void SolarSystemState::mouseMove(const TwoFloats& pos)
{
	int mousePlanet;

	mousePos = screenPos + pos;

	if(!player.getShipIsMoving() &&
	   !gameOver &&
	   (globalData->getNumFleetsDefeated() < 6) &&
	   !displayInfo)
	{
		mousePlanet = getPlanetUnderMouse();
		if(mousePlanet > -1)
			playerTrajectory = planets[mousePlanet]->getPosition();
		else
			playerTrajectory = player.getPosition();
	}

	infoBoxes[currentInfoBox]->mouseMove(pos);
}

//------------------------------------------------------------------------------
int SolarSystemState::getPlanetUnderMouse()
{
	unsigned int i;
	int retval = -1;

	for(i=0;i<planets.size();++i)
	{
		if(mousePos.getDistance(planets[i]->getPosition()) < 32.0f)
		{
			retval = i;
			break;
		}
	}

	return retval;
}

//------------------------------------------------------------------------------
void SolarSystemState::planetFleetCollision()
{
	int i;
	unsigned int j;

	for(i=(planets.size()-1);i>=0;--i)
	{
		for(j=0;j<fleets.size();++j)
		{
			if(planets[i]->getPosition().getDistance(fleets[j]->getPosition()) < 32.0f)
			{
				if(planets[i]->incrementHealth(-0.002f))
				{
					particles.trigger(planets[i]->getPosition());
					planets.erase(planets.begin()+i);
					planetOffsets.erase(planetOffsets.begin()+i);

					shake += 1.0f;

					if(!planets.size())
					{
						gameOver = true;
						timePaused = true;
					}
					else
						updateFleetDestinations();

					break;
				}
			}
		}
	}
}

//------------------------------------------------------------------------------
void SolarSystemState::updateFleetDestinations()
{
	float tempf;
	unsigned int i, j;

	for(i=0;i<fleets.size();++i)
	{
		int nearestIndex = -1;
		float distance = 10000.0f;

		//Tell fleet to move to nearest other planet.
		for(j=0;j<planets.size();++j)
		{
			tempf = planets[j]->getPosition().getDistance(fleets[i]->getPosition());
			if(tempf < distance)
			{
				nearestIndex = j;
				distance = tempf;
			}
		}

		if(nearestIndex > -1)
			fleets[i]->setDestination(planets[nearestIndex]->getPosition());
	}
}

//------------------------------------------------------------------------------
int SolarSystemState::getPlayerPlanet()
{
	float tempf;
	unsigned int i;
	int retval = -1;
	int nearest = -1;
	float dist = 10000.0f;
	const TwoFloats& playerPos = player.getPosition();

	for(i=0;i<planets.size();++i)
	{
		tempf = playerPos.getDistance(planets[i]->getPosition());
		if((tempf < 4.0f) && (tempf < dist))
		{
			nearest = i;
			dist = tempf;
		}
	}
	retval = nearest;

	return retval;
}
