//	SolarSystemState.h - The solar system overview state.
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2012  Niall Moody
//
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef SOLARSYSTEMSTATE_H_
#define SOLARSYSTEMSTATE_H_

#include "synth/BrushStrokesSynth.h"
#include "drawing/TextLayout.h"
#include "ParticleGenerator.h"
#include "TrajectoryLine.h"
#include "Array2DDrawer.h"
#include "PlayerShip.h"
#include "GameState.h"
#include "Sun.h"

class BlobVaderSounder;
class BlobGlobalData;
class InfoBox;
class Planet;
class Fleet;

///	The solar system overview state.
class SolarSystemState : public GameState
{
  public:
	///	Constructor.
	SolarSystemState(Application& app, SounderPtr& sound);
	///	Destructor.
	~SolarSystemState();

	///	Called frequently to update the state.
	void update();
	///	Called to draw the state.
	void draw(DrawerPtr& d);

	///	Called when a key down event is received.
	void keyDown(const std::wstring& action,
				 const std::wstring& keyPressed,
				 int keyModifier);

	///	Called when a mouse down event is received.
	void mouseDown(int button, const TwoFloats& pos);
	///	Used to test connected line drawing.
	void mouseMove(const TwoFloats& pos);

	///	Returns the name of this state.
	const std::wstring getName() const {return L"SolarSystemState";};
	///	Used to construct instances of this state.
	static GameStateCreator<SolarSystemState> creator;
  private:
	///	Helper method. Returns the index of the planet the mouse is over.
	/*!
		\return -1 if the mouse is not over a planet.
	 */
	int getPlanetUnderMouse();
	///	Helper method. Decrements planets' health when they are in contact with a fleet.
	void planetFleetCollision();
	///	Helper method. Updates fleets' destinations when planets are destroyed.
	void updateFleetDestinations();
	///	Helper method. Returns the index of the planet the player is currently orbiting.
	/*!
		\return -1 If the player is not orbiting a planet.
	 */
	int getPlayerPlanet();

	///	The maximum number of planets to generate.
	enum
	{
		NumPlanets = 6,

		PlanetBox = 0,
		EncyclopediaBox,

		NumInfoBoxes
	};

	///	The current background colour.
	ThreeFloats backgroundColour;
	///	Tweener for the background colour.
	STweener<ThreeFloats> backgroundTweener;
	///	True if the background can flash.
	bool backgroundFlash;

	///	The sun.
	Sun sun;
	///	The planets.
	std::vector<shared_ptr<Planet> > planets;
	///	Used to spin the planet trajectories when the mouse is over them.
	std::vector<float> planetOffsets;

	///	The player's ship.
	PlayerShip player;
	///	The BlobVader fleets.
	std::vector<shared_ptr<Fleet> > fleets;

	///	Used to draw the player's trajectory.
	TrajectoryLine playerTrajectoryLine;
	///	Used to draw the fleets' trajectories.
	TrajectoryLine fleetTrajectoryLines[2];
	
	///	Used to shake the screen.
	float shake;
	///	Used to shake the screen.
	float shakeIndex;
	///	True if the screen can shake.
	bool screenShake;
	///	Particles drawn when a planet kerplodes.
	ParticleGenerator particles;

	///	The top left of the screen.
	TwoFloats screenPos;

	///	The days elapsed since the game was started.
	float daysElapsed;
	///	Whether time is currently paused or not.
	bool timePaused;

	///	The current mouse position.
	TwoFloats mousePos;

	///	Mouse trail used by brush strokes.
	TwoFloats mouseTrail[BrushStrokesSynth::TableSize];

	///	The current end point of the player's trajectory.
	TwoFloats playerTrajectory;
	///	The time it would take to get to the end point of the player's trajectory.
	float playerTrajectoryTime;

	///	Used to display the info box.
	bool displayInfo;
	///	The top left position of the info box.
	TwoFloats infoPos;
	///	The size of the info box.
	const TwoFloats infoSize;
	///	The top left position of the close info box button.
	TwoFloats infoClosePos;
	///	The info box TextLayouts.
	TextLayout infoLayouts[2];

	///	The current InfoBox.
	int currentInfoBox;
	///	The available InfoBoxes.
	shared_ptr<InfoBox> infoBoxes[NumInfoBoxes];

	///	The global data for the game.
	shared_ptr<BlobGlobalData> globalData;

	///	True if all the player's planets have been destroyed.
	bool gameOver;
	///	Used to draw the game over text.
	TextLayout gameOverLayout;

	///	Used to draw the game win text.
	TextLayout gameWinLayout;

	///	The play button array.
	Array2D<bool, 5, 5> playButton;
	///	The pause button array.
	Array2D<bool, 5, 5> pauseButton;
	///	The encyclopedia button array.
	Array2D<bool, 5, 5> encyclopediaButton;
	///	Used to draw the various buttons.
	Array2DDrawer<5, 5> arrayDrawer;

	///	Our copy of the sounder.
	shared_ptr<BlobVaderSounder> sounder;
};

#endif
