//	Vader.cpp - Represents a basic BlobVader enemy (new version).
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2012  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "VaderDamageAction.h"
#include "NewGamePlayState.h"
#include "VaderProjectile.h"
#include "VaderKillAction.h"
#include "drawing/Drawer.h"
#include "Random.h"
#include "Vader.h"

//------------------------------------------------------------------------------
Vader::Vader(const std::wstring& vaderName,
			 const TwoFloats& startPos,
			 const Array2D<bool, 7, 7>& blobs,
			 const std::vector<shared_ptr<VaderProjectileCreatorBase> >& vaderProjectiles,
			 const std::vector<shared_ptr<VaderDamageAction> >& vaderDamageActions,
			 const shared_ptr<VaderKillAction>& vaderKillAction,
			 NewGamePlayState& gamePlayState,
			 bool skipInitialBounce):
state(gamePlayState),
position(startPos),
blobsArray(blobs),
name(vaderName),
killAction(vaderKillAction),
scale(0.625f),
scaleTween(scale, scale, 0.075f),
initialTween(position, TwoFloats(startPos.x, startPos.y+448.0f), 0.01f),
stepTween(position, TwoFloats(startPos), 0.05f),
projectileInc(1),
dead(false)
{
	unsigned int i;

	wingMotion = Random::getGlobal().getRandomFloat();
	initialTween.setDelay(Random::getGlobal().getRandomFloat());
	stepTween.setIndex(1.0f);

	if(skipInitialBounce)
	{
		initialTween.reset(startPos);
		initialTween.setIndex(1.0f);
	}

	for(i=0;i<vaderProjectiles.size();++i)
		projectiles.push_back(vaderProjectiles[i]);
	for(i=0;i<vaderDamageActions.size();++i)
		damageActions.push_back(vaderDamageActions[i]);

	projectileCount = 240 + Random::getGlobal().getRandomInt(480);
}

//------------------------------------------------------------------------------
Vader::~Vader()
{

}

//------------------------------------------------------------------------------
void Vader::update()
{
	float tempf;
	unsigned int i;
	TwoFloats tempPos;

	if(!initialTween.getFinished())
		initialTween.tick();
	else if(!stepTween.getFinished())
		stepTween.tick();

	if(!scaleTween.getFinished())
		scaleTween.tick();

	projectileCount -= projectileInc;
	if(!projectileCount)
	{
		tempPos = position;
		if(projectiles.size() > 1)
			tempPos.x -= 45.0f;
		for(i=0;i<projectiles.size();++i)
		{
			state.addVaderProjectile(projectiles[i]->createInstance(tempPos));

			tempPos.x += 90.0f/(float)(projectiles.size()-1);
		}

		if(projectiles.size() > 0)
		{
			scale = 1.25f;
			scaleTween.reset(0.625f);
		}

		tempf = state.getTimeElapsed()/60.0f;
		if(tempf > 1.0f)
			tempf = 1.0f;
		tempf = 1.0f-tempf;

		projectileCount = 120 + (int)(tempf * 120.0f);
		projectileCount += Random::getGlobal().getRandomInt(120 + (int)(tempf * 420.0f));
	}
}

//------------------------------------------------------------------------------
void Vader::draw(shared_ptr<Drawer>& d,
				 const ThreeFloats& leftCol,
				 const ThreeFloats& rightCol)
{
	float tempf;
	unsigned int x, y;
	TwoFloats tempPos;
	TwoFloats tempDir;

	tempf = (position.x - 128.0f)/(1280.0f-256.0f);
	colour = (leftCol * (1.0f-tempf)) + (rightCol * tempf);

#ifdef DEBUG
	d->drawImage(L"Blob", getPosition()-32.0f, TwoFloats(2.0f), colour);
#else
	tempf = sinf(wingMotion*2.0f*NIALL_PI);
	for(y=0;y<7;++y)
	{
		for(x=0;x<7;++x)
		{
			if(blobsArray[y][x])
			{
				//i.e.	  centre   - top left of blob image     - top left of blob array
				tempPos = position - (TwoFloats(16.0f * scale)) - 30.0f;
				tempPos.x += 10.0f * (float)x;
				tempPos.y += 10.0f * (float)y;

				//Add in wing motion.
				if(x < 3)
					tempPos.y += (tempf * ((float)(3-x)/3.0f)) * 8.0f;
				else if(x > 3)
					tempPos.y += (tempf * ((float)(x-3)/3.0f)) * 8.0f;

				d->drawImage(L"Blob", tempPos, TwoFloats(scale), colour);
			}
		}
	}

	wingMotion += 0.01f;
	if(wingMotion > 1.0f)
		wingMotion -= 1.0f;
#endif
}

//------------------------------------------------------------------------------
void Vader::stepLeft()
{
	TwoFloats tempPos(position);

	tempPos.x -= (1280.0f-256.0f)/8.0f;
	stepTween.setDelay(Random::getGlobal().getRandomFloat() * 0.5f);
	stepTween.reset(tempPos);
}

//------------------------------------------------------------------------------
void Vader::stepRight()
{
	TwoFloats tempPos(position);

	tempPos.x += (1280.0f-256.0f)/8.0f;
	stepTween.setDelay(Random::getGlobal().getRandomFloat() * 0.5f);
	stepTween.reset(tempPos);
}

//------------------------------------------------------------------------------
void Vader::stepDown()
{
	TwoFloats tempPos(position);

	tempPos.y += (640-256.0f)/4.0f;
	stepTween.setDelay(Random::getGlobal().getRandomFloat() * 0.5f);
	stepTween.reset(tempPos);
}

//------------------------------------------------------------------------------
void Vader::scaleUp()
{
	scale = 1.25f;
	scaleTween.reset(0.625f);
}

//------------------------------------------------------------------------------
void Vader::setPosition(const TwoFloats& pos)
{
	position = pos;
	initialTween.reset(pos);
	initialTween.setIndex(1.0f);
	stepTween.reset(pos);
	stepTween.setIndex(1.0f);
}

//------------------------------------------------------------------------------
void Vader::setBulletSpeed(int val)
{
	projectileInc = val;
}

//------------------------------------------------------------------------------
void Vader::collided(const std::wstring& other, const TwoFloats& otherPos)
{
	unsigned int i;

	if((other == L"Player") ||
	   (other == L"PlayerBullet") ||
	   (other == L"VaderBulletOwn"))
	{
		if(!damageActions.size())
			setDead(true);
		else
		{
			for(i=0;i<damageActions.size();++i)
				damageActions[i]->damageTaken(*this, state);
		}
	}
}

//------------------------------------------------------------------------------
void Vader::setDead(bool val)
{
	dead = val;

	if(dead)
	{
		if(killAction)
			killAction->killed(*this, state);
	}
}
