//	Vader.h - Represents a basic BlobVader enemy (new version).
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2012  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef VADER_H_
#define VADER_H_

#include "NewGamePlayState.h"
#include "ThreeFloats.h"
#include "c0xHeaders.h"
#include "Collidable.h"
#include "TwoFloats.h"
#include "Tweener.h"
#include "Array2D.h"

#include <vector>

class VaderProjectileCreatorBase;
class VaderDamageAction;
class VaderKillAction;
class Drawer;

///	Represents a basic BlobVader enemy (new version).
class Vader : public Collidable
{
  public:
	///	Constructor.
	/*!
		\param vaderName The vader's name.
		\param startPos The vader's initial position.
		\param blobs The 1-bit bitmap which defines the vader's shape.
		\param projectiles The projectiles this vader fires.
		\param damageActions The actions this vader takes when it is damaged.
		\param killAction The action this vader takes when it is killed.
		\param gamePlayState The owning state.
		\param skipInitialBounce If true, the vader won't do the initial bounce
		down to its start position.
	 */
	Vader(const std::wstring& vaderName,
		  const TwoFloats& startPos,
		  const Array2D<bool, 7, 7>& blobs,
		  const std::vector<shared_ptr<VaderProjectileCreatorBase> >& vaderProjectiles,
		  const std::vector<shared_ptr<VaderDamageAction> >& vaderDamageActions,
		  const shared_ptr<VaderKillAction>& vaderKillAction,
		  NewGamePlayState& gamePlayState,
		  bool skipInitialBounce = false);
	///	Destructor.
	~Vader();

	///	Updates the vader.
	void update();
	///	Draws the vader.
	void draw(shared_ptr<Drawer>& d,
			  const ThreeFloats& leftCol,
			  const ThreeFloats& rightCol);

	///	Sets the BlobVader to step left.
	void stepLeft();
	///	Sets the BlobVader to step right.
	void stepRight();
	///	Sets the BlobVader to step down.
	void stepDown();

	///	Tells the BlobVader to briefly scale up.
	void scaleUp();

	///	Sets the vader's position.
	void setPosition(const TwoFloats& pos);
	///	Updates how fast the vader fires.
	void setBulletSpeed(int val);

	///	Returns the actor's collision radius.
	float getRadius() const {return 32.0f;};
	///	Returns the actor's position.
	const TwoFloats& getPosition() const {return position;};

	///	Returns the actor's type name.
	const std::wstring getName() const {return L"BlobVader";};
	///	Called when the actor has collided with another.
	void collided(const std::wstring& other, const TwoFloats& otherPos);
	///	Returns true if this actor should not be counted in collisions.
	bool doNotCount() const {return !initialTween.getFinished();};

	///	Returns the vader's name.
	const std::wstring& getVaderName() const {return name;};
	///	Returns the vader's image blobs so they can be modified.
	Array2D<bool, 7, 7>& getImage() {return blobsArray;};
	///	Returns the vader's current colour.
	const ThreeFloats& getColour() const {return colour;};
	///	Returns the vader's current wing motion.
	float getWingMotion() const {return wingMotion;};
	///	Returns the vader's projectile creators.
	const std::vector<shared_ptr<VaderProjectileCreatorBase> >& getProjectileCreators() const {return projectiles;};
	///	Returns the vader's damage actions.
	const std::vector<shared_ptr<VaderDamageAction> >& getDamageActions() const {return damageActions;};
	///	Returns the vader's kill actions.
	const shared_ptr<VaderKillAction>& getKillAction() const {return killAction;};

	///	Sets the vader's dead variable.
	void setDead(bool val);
	///	Returns true if the vader should be deleted.
	bool getDead() const {return dead;};
  private:
	///	The owning NewGamePlayState.
	NewGamePlayState& state;

	///	The vader's position.
	TwoFloats position;
	///	The vader's image.
	Array2D<bool, 7, 7> blobsArray;
	///	Used to make the vader's 'wings' move.
	float wingMotion;
	///	The vader's current colour.
	ThreeFloats colour;

	///	The vader's name.
	std::wstring name;
	///	The vader's projectiles.
	std::vector<shared_ptr<VaderProjectileCreatorBase> > projectiles;
	///	The vader's damage actions.
	std::vector<shared_ptr<VaderDamageAction> > damageActions;
	///	The vader's kill action.
	shared_ptr<VaderKillAction> killAction;

	///	The scale of the vader's blobs.
	float scale;
	///	Used to tween the scale of the vader's blobs.
	BounceTweener<float> scaleTween;

	///	Used to tween the vader down from the top of the screen.
	BounceTweener<TwoFloats> initialTween;
	///	Used to tween the vader to it's next position (via stepLeft() etc.).
	STweener<TwoFloats> stepTween;

	///	The counter used to determine when to fire another projectile.
	int projectileCount;
	///	Used to determine how fast to add to projectileCount.
	int projectileInc;

	///	True if it's time to delete the vader.
	bool dead;
};

#endif
