//	VaderFactory.cpp - Class used to construct procedural Vaders.
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2012  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "VaderDamageAction.h"
#include "VaderProjectile.h"
#include "VaderKillAction.h"
#include "VaderFactory.h"
#include "Random.h"
#include "Vader.h"

#include <algorithm>

//------------------------------------------------------------------------------
VaderFactory::VaderFactory()
{

}

//------------------------------------------------------------------------------
VaderFactory::~VaderFactory()
{

}

//------------------------------------------------------------------------------
void VaderFactory::initialise()
{
	int i, j, k;
	int numTypes;
	int difficulty;
	bool alreadyExists;
	VaderType tempType;
	std::multimap<int, VaderType>::iterator it;

	//Clear the previous vaderTypes.
	vaderTypes.clear();

	//Add the absolute basic vader (i.e. no projectiles or damage actions) first.
	tempType.name = createRandomName();
	createRandomImage(tempType.image);
	vaderTypes.insert(std::make_pair(0, tempType));

	//numTypes = projectiles.size() + damageActions.size();
	numTypes = 3;

	//For each difficulty level.
	for(difficulty=1;difficulty<5;++difficulty)
	{
		//Create 'numTypes' VaderTypes.
		for(i=0;i<numTypes;++i)
		{
			alreadyExists = true;

			tempType.name = createRandomName();
			createRandomImage(tempType.image);

			while(alreadyExists)
			{
				std::multimap<int, shared_ptr<VaderDamageActionCreatorBase> >::iterator it2;
				std::multimap<int, shared_ptr<VaderKillActionCreatorBase> >::iterator it3;
				std::vector<shared_ptr<VaderDamageActionCreatorBase> > availableActions;
				std::vector<shared_ptr<VaderKillActionCreatorBase> > availableKills;

				//Add all the relevant damage actions to a vector from which they will be
				//removed once they're added to tempType.
				for(it2=damageActions.lower_bound(0);
					it2!=damageActions.upper_bound(difficulty);
					++it2)
				{
					availableActions.push_back(it2->second);
				}

				//Ditto for killActions
				for(it3=killActions.lower_bound(0);
					it3!=killActions.upper_bound(difficulty);
					++it3)
				{
					availableKills.push_back(it3->second);
				}

				alreadyExists = false;

				tempType.projectiles.clear();
				tempType.damageActions.clear();
				tempType.killAction.reset();

				for(j=0;j<difficulty;)
				{
					//Remove any damage or kill actions beyond the remaining
					//difficulty limit.
					for(k=(availableActions.size()-1);k>=0;--k)
					{
						if(availableActions[k]->getDifficulty() > (difficulty-j))
							availableActions.erase(availableActions.begin()+k);
					}
					for(k=(availableKills.size()-1);k>=0;--k)
					{
						if(availableKills[k]->getDifficulty() > (difficulty-j))
							availableKills.erase(availableKills.begin()+k);
					}

					if(!tempType.killAction &&
					   availableKills.size() &&
					   (Random::getGlobal().getRandomFloat() > 0.66f))
					{
						tempType.killAction = availableKills[Random::getGlobal().getRandomInt(availableKills.size())];
						j += tempType.killAction->getDifficulty();
					}
					else if(availableActions.size() > 0)
					{
						//Add random projectile to tempType.
						if(Random::getGlobal().getRandomFloat() > 0.5f)
						{
							tempType.projectiles.push_back(getRandomProjectile(difficulty-j));
							j += tempType.projectiles.back()->getDifficulty();
						}
						else //Add random damage action to tempType.
						{
							tempType.damageActions.push_back(getRandomDamageAction(availableActions));
							j += tempType.damageActions.back()->getDifficulty();
						}
					}
					else //Add random projectile to tempType.
					{
						tempType.projectiles.push_back(getRandomProjectile(difficulty));
						j += tempType.projectiles.back()->getDifficulty();
					}
				}

				//Sort tempType's two vectors.
				std::sort(tempType.projectiles.begin(),
						  tempType.projectiles.end(),
						  projectileCmp);
				std::sort(tempType.damageActions.begin(),
						  tempType.damageActions.end(),
						  damageActionCmp);

				//Check if this combination of projectile and/or damage action matches an existing type.
				for(it=vaderTypes.lower_bound(difficulty);
					it!=vaderTypes.upper_bound(difficulty);
					++it)
				{
					//If tempType matches it->second...
					if((tempType.projectiles.size() == it->second.projectiles.size()) &&
					   (tempType.damageActions.size() == it->second.damageActions.size()) &&
					   (tempType.killAction == it->second.killAction))
					{
						if((std::equal(tempType.projectiles.begin(),
									   tempType.projectiles.end(),
									   it->second.projectiles.begin())) &&
						   (std::equal(tempType.damageActions.begin(),
									   tempType.damageActions.end(),
									   it->second.damageActions.begin())))
						{
							alreadyExists = true;
							break;
						}
					}
				}
			}

			vaderTypes.insert(std::make_pair(difficulty, tempType));
		}

		//First 2 difficulty levels have the most VaderTypes.
		/*if(difficulty > 1)
			--numTypes;*/
	}
}

//------------------------------------------------------------------------------
shared_ptr<Vader> VaderFactory::createVader(const std::wstring& name,
											const TwoFloats& pos,
											NewGamePlayState& state) const
{
	unsigned int i;
	std::multimap<int, VaderType>::const_iterator it;

	for(it=vaderTypes.begin();it!=vaderTypes.end();++it)
	{
		if(it->second.name == name)
		{
			shared_ptr<VaderKillAction> killAction;
			std::vector<shared_ptr<VaderDamageAction> > tempActions;

			for(i=0;i<it->second.damageActions.size();++i)
				tempActions.push_back(it->second.damageActions[i]->createInstance());

			if(it->second.killAction)
				killAction = it->second.killAction->createInstance();

			return shared_ptr<Vader>(new Vader(name,
											   pos,
											   it->second.image,
											   it->second.projectiles,
											   tempActions,
											   killAction,
											   state));
		}
	}

	//If we get here we couldn't find the named vader, so just return the
	//difficulty 0 one.
	{
		shared_ptr<VaderKillAction> killAction;
		std::vector<shared_ptr<VaderDamageAction> > tempActions;

		it = vaderTypes.begin();

		for(i=0;i<it->second.damageActions.size();++i)
			tempActions.push_back(it->second.damageActions[i]->createInstance());

			if(it->second.killAction)
				killAction = it->second.killAction->createInstance();

		return shared_ptr<Vader>(new Vader(name,
										   pos,
										   it->second.image,
										   it->second.projectiles,
										   tempActions,
										   killAction,
										   state));
	}
}

//------------------------------------------------------------------------------
const std::wstring VaderFactory::getVaderName(int difficulty, int index) const
{
	int i;
	std::wstring retval;
	std::multimap<int, VaderType>::const_iterator it;

	while(retval == L"")
	{
		for(it=vaderTypes.begin(), i=0;
			it!=vaderTypes.end();
			++it, ++i)
		{
			if(it->first > difficulty)
			{
				index %= i;
				break;
			}
			if(i == index)
			{
				retval = it->second.name;
				break;
			}
		}
	}

	return retval;
}

//------------------------------------------------------------------------------
bool VaderFactory::projectileCmp(const shared_ptr<VaderProjectileCreatorBase>& left,
							     const shared_ptr<VaderProjectileCreatorBase>& right)
{
	return (left->getDifficulty() < right->getDifficulty());
}

//------------------------------------------------------------------------------
bool VaderFactory::damageActionCmp(const shared_ptr<VaderDamageActionCreatorBase>& left,
								   const shared_ptr<VaderDamageActionCreatorBase>& right)
{
	return (left->getDifficulty() < right->getDifficulty());
}

//------------------------------------------------------------------------------
const std::wstring VaderFactory::createRandomName()
{
	int index;
	std::wstring retval;
	bool alreadyExists = true;
	std::multimap<int, VaderType>::iterator it;
	const std::wstring first[] = {L"Ro",
								  L"Ma",
								  L"Sa",
								  L"Re",
								  L"Le",
								  L"Pe",
								  L"Ba",
								  L"Te",
								  L"Vu",
								  L"Hu",
								  L"Ca",
								  L"Ju",
								  L"Mo",
								  L"Ku",
								  L"Wa",
								  L"Do"};
	const std::wstring second[] = {L"ma",
								   L"ju",
								   L"ye",
								   L"nu",
								   L"zi",
								   L"se",
								   L"lo",
								   L"ca",
								   L"be",
								   L"yi",
								   L"fe",
								   L"qu",
								   L"no",
								   L"xa",
								   L"ta",
								   L"vu"};
	const std::wstring third[] = {L"tron",
								  L"ma",
								  L"zig",
								  L"tu",
								  L"ra",
								  L"za",
								  L"rine",
								  L"me",
								  L"lia",
								  L"ny",
								  L" 3000",
								  L"lion",
								  L"ly",
								  L"sion",
								  L"rah",
								  L"zer"};

	while(alreadyExists)
	{
		index = Random::getGlobal().getRandomInt(16);
		index = (index > 15) ? 15 : index;
		retval = first[index];

		index = Random::getGlobal().getRandomInt(16);
		index = (index > 15) ? 15 : index;
		retval.append(second[index]);

		index = Random::getGlobal().getRandomInt(16);
		index = (index > 15) ? 15 : index;
		retval.append(third[index]);

		alreadyExists = false;
		for(it=vaderTypes.begin();it!=vaderTypes.end();++it)
		{
			if(it->second.name == retval)
			{
				alreadyExists = true;
				break;
			}
		}
	}

	return retval;
}

//------------------------------------------------------------------------------
void VaderFactory::createRandomImage(Array2D<bool, 7, 7>& image)
{
	int x, y;
	int blobCount = 0;
	float chance = 0.75f;

	for(y=0;y<7;++y)
	{
		for(x=0;x<7;++x)
			image[y][x] = false;
	}

	image[3][3] = true;
	++blobCount;

	x = 3;
	for(y=2;y>=0;--y)
	{
		if(Random::getGlobal().getRandomFloat() > chance)
			break;
		chance -= 0.25f;

		image[y][x] = true;
		++blobCount;
	}
	for(y=4;y<7;++y)
	{
		if(Random::getGlobal().getRandomFloat() > chance)
			break;
		chance -= 0.25f;

		image[y][x] = true;
		++blobCount;
	}

	while(blobCount < 16)
	{
		for(x=2;x>=0;--x)
		{
			for(y=2;y>=0;--y)
			{
				if(image[y][x])
					continue;

				if((image[y][x+1] && (Random::getGlobal().getRandomFloat() > 0.5f)) ||
				   (Random::getGlobal().getRandomFloat() > 0.75f))
				{
					if((x < 2) && !image[y][x+1] && !image[y+1][x])
						continue;

					image[y][x] = true;
					++blobCount;

					image[y][6-x] = true;
					++blobCount;
				}
			}

			for(y=4;y<7;++y)
			{
				if(image[y][x])
					continue;

				if((image[y][x+1] && (Random::getGlobal().getRandomFloat() > 0.5f)) ||
				   (Random::getGlobal().getRandomFloat() > 0.75f))
				{
					if((x < 2) && !image[y][x+1] && !image[y-1][x])
						continue;

					image[y][x] = true;
					++blobCount;

					image[y][6-x] = true;
					++blobCount;
				}
			}
		}
	}
}

//------------------------------------------------------------------------------
shared_ptr<VaderProjectileCreatorBase>& VaderFactory::getRandomProjectile(int difficulty)
{
	int i;
	int index = Random::getGlobal().getRandomInt(projectiles.size());
	std::multimap<int, shared_ptr<VaderProjectileCreatorBase> >::iterator it;

	while(1)
	{
		for(it=projectiles.begin(), i=0;
			it!=projectiles.end();
			++it, ++i)
		{
			if(it->first > difficulty)
			{
				index %= i;
				break;
			}
			if(i == index)
				return it->second;
		}
	}
}

//------------------------------------------------------------------------------
shared_ptr<VaderDamageActionCreatorBase> VaderFactory::getRandomDamageAction(std::vector<shared_ptr<VaderDamageActionCreatorBase> >& availableActions)
{
	int index = Random::getGlobal().getRandomInt(availableActions.size());
	shared_ptr<VaderDamageActionCreatorBase> retval;

	index = Random::getGlobal().getRandomInt(availableActions.size());

	retval = availableActions[index];
	availableActions.erase(availableActions.begin()+index);

	return retval;
}
