//	VaderFactory.h - Class used to construct procedural Vaders.
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2012  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef VADERFACTORY_H_
#define VADERFACTORY_H_

#include "VaderDamageAction.h"
#include "VaderProjectile.h"
#include "VaderKillAction.h"
#include "c0xHeaders.h"
#include "TwoFloats.h"
#include "Array2D.h"

#include <vector>
#include <string>
#include <map>

//class VaderDamageActionCreatorBase;
//class VaderProjectileCreatorBase;
//class VaderKillActionCreatorBase;
class NewGamePlayState;
class Vader;

///	Class used to construct procedural Vaders.
class VaderFactory
{
  public:
	///	Constructor.
	VaderFactory();
	///	Destructor.
	~VaderFactory();

	///	Registers a VaderProjectileCreator with the factory.
	template<class T>
	void registerProjectile(int difficulty,
							const std::wstring& name,
							const std::wstring& desc)
	{
		projectiles.insert(std::make_pair(difficulty,
										  shared_ptr<VaderProjectileCreatorBase>(new VaderProjectileCreator<T>(difficulty,
																											   name,
																											   desc))));
	};
	///	Registers a VaderDamageActionCreator with the factory.
	template<class T>
	void registerDamageAction(int difficulty)
	{
		damageActions.insert(std::make_pair(difficulty,
											shared_ptr<VaderDamageActionCreatorBase>(new VaderDamageActionCreator<T>(difficulty))));
	}
	///	Registers a VaderKillActionCreator with the factory.
	template<class T>
	void registerKillAction(int difficulty)
	{
		killActions.insert(std::make_pair(difficulty,
										  shared_ptr<VaderKillActionCreatorBase>(new VaderKillActionCreator<T>(difficulty))));
	}

	///	Initialises the factory.
	/*!
		Make sure you've registered all your VaderProjectiles and
		VaderDamageActions before you call this.

		This creates all the VaderTypes for a single game.
	 */
	void initialise();

	///	Creates the named Vader at the passed-in position.
	shared_ptr<Vader> createVader(const std::wstring& name,
								  const TwoFloats& pos,
								  NewGamePlayState& state) const;

	///	Returns the total number of Vader types.
	int getNumVaderTypes() const {return vaderTypes.size();};
	///	Returns the name of the indexed VaderType up to the specified difficulty.
	/*!
		If index points to a VaderType of difficulty higher than that passed in,
		it will be wrapped around modulo-style so you are guaranteed a VaderType
		within your desired difficulty scale.
	 */
	const std::wstring getVaderName(int difficulty, int index) const;

	///	Helper method used to sort VaderType::projectiles.
	static bool projectileCmp(const shared_ptr<VaderProjectileCreatorBase>& left,
							  const shared_ptr<VaderProjectileCreatorBase>& right);
	///	Helper method used to sort VaderType::damageActions.
	static bool damageActionCmp(const shared_ptr<VaderDamageActionCreatorBase>& left,
								const shared_ptr<VaderDamageActionCreatorBase>& right);
  private:
	///	Generates a random name for a vader.
	const std::wstring createRandomName();
	///	Generates a random image for a vader.
	void createRandomImage(Array2D<bool, 7, 7>& image);

	///	Helper method. Returns a random ProjectileCreator for the specified difficulty.
	shared_ptr<VaderProjectileCreatorBase>& getRandomProjectile(int difficulty);
	///	Helper method. Returns a random DamageAction for the specified difficulty.
	shared_ptr<VaderDamageActionCreatorBase> getRandomDamageAction(std::vector<shared_ptr<VaderDamageActionCreatorBase> >& availableActions);

	///	Struct representing a particular type of Vader.
	struct VaderType
	{
		///	The type's name.
		std::wstring name;
		///	The vader's image.
		Array2D<bool, 7, 7> image;
		///	The projectiles this vader fires.
		std::vector<shared_ptr<VaderProjectileCreatorBase> > projectiles;
		///	The damage actions this vader has.
		std::vector<shared_ptr<VaderDamageActionCreatorBase> > damageActions;
		///	The vader's kill action.
		shared_ptr<VaderKillActionCreatorBase> killAction;
	};

	///	All the generated vader types.
	/*!
		The int is the difficulty of this type.
	 */
	std::multimap<int, VaderType> vaderTypes;

	///	The registered VaderProjectiles.
	/*!
		The int is the difficulty of this projectile.
	 */
	std::multimap<int, shared_ptr<VaderProjectileCreatorBase> > projectiles;
	///	The registered VaderDamageTypes.
	/*!
		The int is the difficulty of this damage type.
	 */
	std::multimap<int, shared_ptr<VaderDamageActionCreatorBase> > damageActions;
	///	The registered VaderKillTypes.
	/*!
		The int is the difficulty of this kill type.
	 */
	std::multimap<int, shared_ptr<VaderKillActionCreatorBase> > killActions;
};

#endif
